import logging
import socket
import time
from threading import Thread

from autouri import AutoURI

logger = logging.getLogger(__name__)


class ServerHeartbeatTimeoutError(Exception):
    pass


class ServerHeartbeat:
    DEFAULT_SERVER_HEARTBEAT_FILE = '~/.caper/default_server_heartbeat'
    DEFAULT_HEARTBEAT_TIMEOUT_MS = 120000
    DEFAULT_INTERVAL_UPDATE_HEARTBEAT_SEC = 60.0

    def __init__(
        self,
        heartbeat_file=DEFAULT_SERVER_HEARTBEAT_FILE,
        heartbeat_timeout=DEFAULT_HEARTBEAT_TIMEOUT_MS,
        interval_update_heartbeat=DEFAULT_INTERVAL_UPDATE_HEARTBEAT_SEC,
    ):
        """Server heartbeat to share store server's hostname/port with clients.

        Args:
            heartbeat_file:
                Server writes hostname/port on this file.
                Client reads hostname/port from this file.
            heartbeat_timeout:
                Expiration period for a heartbeat file (in milliseconds).
                Client will use a heartbeat file only if it is fresh (within timeout).
            interval_update_heartbeat:
                Period for updtaing a heartbeat file (in seconds).
        """
        self._heartbeat_file = heartbeat_file
        self._heartbeat_timeout = heartbeat_timeout
        self._interval_update_heartbeat = interval_update_heartbeat

        self._stop_it = False
        self._thread = None

    def start(self, port, hostname=None):
        """Starts a thread that writes hostname/port of a server
        on a heartbeat file.

        Args:
            port:
                This port will be written to a heartbeat file.
            hostname:
                Optional hostname to be written to heartbeat file.
                socket.gethostname() will be used if not defined.
        """
        self._thread = Thread(target=self._write_to_file, args=(port, hostname))
        self._thread.start()
        return self._thread

    def is_alive(self):
        return self._thread.is_alive() if self._thread else False

    def stop(self):
        self._stop_it = True

        if self._thread:
            self._thread.join()

    def read(self, raise_timeout=False):
        """Read from heartbeat file.
        If a heartbeat file is not fresh (mtime difference < timeout)
        then None is returned.

        Returns:
            Tuple of (hostname, port)
        """
        try:
            u = AutoURI(self._heartbeat_file)
            if (time.time() - u.mtime) * 1000.0 > self._heartbeat_timeout:
                raise ServerHeartbeatTimeoutError
            else:
                hostname, port = u.read().strip('\n').split(':')
                logger.info(
                    'Reading hostname/port from a heartbeat file. {h}:{p}'.format(
                        h=hostname, p=port
                    )
                )
                return hostname, int(port)

        except ServerHeartbeatTimeoutError:
            logger.error(
                'Found a heartbeat file but it has been expired (> timeout)'
                '. {f}'.format(f=self._heartbeat_file)
            )
            if raise_timeout:
                raise

        except Exception:
            logger.error(
                'Failed to read from a heartbeat file. {f}'.format(
                    f=self._heartbeat_file
                )
            )

    def _write_to_file(self, port, hostname=None):
        if not hostname:
            hostname = socket.gethostname()

        logger.info('Server heartbeat thread started.')

        while True:
            try:
                logger.debug(
                    'Writing heartbeat: {hostname}, {port}'.format(
                        hostname=hostname, port=port
                    )
                )
                AutoURI(self._heartbeat_file).write(
                    '{hostname}:{port}'.format(hostname=hostname, port=port)
                )
            except Exception:
                logger.error(
                    'Failed to write to a heartbeat_file. {f}'.format(
                        f=self._heartbeat_file
                    )
                )
            cnt = 0
            while cnt < self._interval_update_heartbeat:
                cnt += 1
                if self._stop_it:
                    break
                time.sleep(1)
            if self._stop_it:
                break

        logger.info('Server heartbeat thread ended.')



import os

from .cromwell import Cromwell
from .cromwell_backend import (
    BACKEND_ALIAS_LOCAL,
    BACKEND_AWS,
    BACKEND_GCP,
    BACKEND_LOCAL,
    BACKEND_PBS,
    BACKEND_SGE,
    BACKEND_SLURM,
)

BACKEND_ALIAS_SHERLOCK = 'sherlock'
BACKEND_ALIAS_SCG = 'scg'

CONF_CONTENTS_LOCAL_HASH_STRAT = """
# Hashing strategy for call-caching (3 choices)
# This parameter is for local (local/slurm/sge/pbs) backend only.
# This is important for call-caching,
# which means re-using outputs from previous/failed workflows.
# Cache will miss if different strategy is used.
# "file" method has been default for all old versions of Caper<1.0.
# "path+modtime" is a new default for Caper>=1.0,
#   file: use md5sum hash (slow).
#   path: use path.
#   path+modtime: use path and modification time.
local-hash-strat=path+modtime
"""

CONF_CONTENTS_TMP_DIR = """
# Local directory for localized files and Cromwell's intermediate files
# If not defined, Caper will make .caper_tmp/ on local-out-dir or CWD.
# /tmp is not recommended here since Caper store all localized data files
# on this directory (e.g. input FASTQs defined as URLs in input JSON).
local-loc-dir=
"""

DEFAULT_CONF_CONTENTS_LOCAL = (
    """
backend=local
"""
    + CONF_CONTENTS_LOCAL_HASH_STRAT
    + CONF_CONTENTS_TMP_DIR
)

DEFAULT_CONF_CONTENTS_SHERLOCK = (
    """
backend=slurm
slurm-partition=

# IMPORTANT warning for Stanford Sherlock cluster
# ====================================================================
# DO NOT install any codes/executables
# (java, conda, python, caper, pipeline's WDL, pipeline's Conda env, ...) on $SCRATCH or $OAK.
# You will see Segmentation Fault errors.
# Install all executables on $HOME or $PI_HOME instead.
# It's STILL OKAY to read input data from and write outputs to $SCRATCH or $OAK.
# ====================================================================
"""
    + CONF_CONTENTS_LOCAL_HASH_STRAT
    + CONF_CONTENTS_TMP_DIR
)

DEFAULT_CONF_CONTENTS_SCG = (
    """
backend=slurm
slurm-account=

"""
    + CONF_CONTENTS_LOCAL_HASH_STRAT
    + CONF_CONTENTS_TMP_DIR
)

DEFAULT_CONF_CONTENTS_SLURM = (
    """
backend=slurm

# define one of the followings (or both) according to your
# cluster's SLURM configuration.
slurm-partition=
slurm-account=
"""
    + CONF_CONTENTS_LOCAL_HASH_STRAT
    + CONF_CONTENTS_TMP_DIR
)

DEFAULT_CONF_CONTENTS_SGE = (
    """
backend=sge
sge-pe=
"""
    + CONF_CONTENTS_LOCAL_HASH_STRAT
    + CONF_CONTENTS_TMP_DIR
)

DEFAULT_CONF_CONTENTS_PBS = (
    """
backend=pbs
"""
    + CONF_CONTENTS_LOCAL_HASH_STRAT
    + CONF_CONTENTS_TMP_DIR
)

DEFAULT_CONF_CONTENTS_AWS = (
    """
backend=aws
aws-batch-arn=
aws-region=
aws-out-dir=
"""
    + CONF_CONTENTS_TMP_DIR
)

DEFAULT_CONF_CONTENTS_GCP = (
    """
backend=gcp
gcp-prj=
gcp-out-dir=

# Call-cached outputs will be duplicated by making a copy or reference
#   reference: refer to old output file in metadata.json file.
#   copy (not recommended): make a copy for a new workflow.
gcp-call-caching-dup-strat=

# Use Google Cloud Life Sciences API instead of Genomics API (deprecating).
# Make sure to enable Google Cloud Life Sciences API on your Google Cloud Console
use-google-cloud-life-sciences=false

# gcp-region is required for Life Sciences API only.
# Region is different from zone. Zone is more specific.
# Do not define zone here. Check supported regions:
#   https://cloud.google.com/life-sciences/docs/concepts/locations
# e.g. us-central1
gcp-region=

# Comma-separated zones for Genomics API (deprecating).
# This is ignored if use-google-cloud-life-sciences.
# e.g. us-west1-a,us-west1-b,us-west1-c
gcp-zones=

# Increase instance's memory when retrying upon OOM (out of memory) error.
gcp-memory-retry-multiplier=1.2

# Number of retrials. This parameter also applies to non-OOM failures.
max-retries=1
"""
    + CONF_CONTENTS_TMP_DIR
)


def init_caper_conf(conf_file, backend):
    """Initialize conf file for a given backend.
    There are two special backend aliases for two Stanford clusters.
    These clusters are based on SLURM.
    Also, download/install Cromwell/Womtool JARs, whose
    default URL and install dir are defined in class Cromwell.
    """
    if backend in (BACKEND_LOCAL, BACKEND_ALIAS_LOCAL):
        contents = DEFAULT_CONF_CONTENTS_LOCAL
    elif backend == BACKEND_ALIAS_SHERLOCK:
        contents = DEFAULT_CONF_CONTENTS_SHERLOCK
    elif backend == BACKEND_ALIAS_SCG:
        contents = DEFAULT_CONF_CONTENTS_SCG
    elif backend == BACKEND_SLURM:
        contents = DEFAULT_CONF_CONTENTS_SLURM
    elif backend == BACKEND_SGE:
        contents = DEFAULT_CONF_CONTENTS_SGE
    elif backend == BACKEND_PBS:
        contents = DEFAULT_CONF_CONTENTS_PBS
    elif backend in BACKEND_GCP:
        contents = DEFAULT_CONF_CONTENTS_GCP
    elif backend in BACKEND_AWS:
        contents = DEFAULT_CONF_CONTENTS_AWS
    else:
        raise ValueError('Unsupported backend {p}'.format(p=backend))

    conf_file = os.path.expanduser(conf_file)
    os.makedirs(os.path.dirname(conf_file), exist_ok=True)

    with open(conf_file, 'w') as fp:
        fp.write(contents + '\n')

        cromwell = Cromwell()
        fp.write(
            '{key}={val}\n'.format(key='cromwell', val=cromwell.install_cromwell())
        )
        fp.write('{key}={val}\n'.format(key='womtool', val=cromwell.install_womtool()))



import hashlib
import json
import logging
import re

from pyhocon import ConfigFactory, HOCONConverter

from .dict_tool import merge_dict

logger = logging.getLogger(__name__)


NEW_LINE = '\n'
RE_HOCON_INCLUDE = [
    r'include\s+(?:required|url|file|classpath)\(.*\)',
    r'include\s+".*\.(?:conf|hocon)"',
]
RE_HOCONSTRING_INCLUDE = r'HOCONSTRING_INCLUDE_(?:.*)\s*=\s*"(?:.*)"'
RE_HOCONSTRING_INCLUDE_VALUE = r'HOCONSTRING_INCLUDE_(?:.*)\s*=\s*"(.*)"'
HOCONSTRING_INCLUDE_KEY = 'HOCONSTRING_INCLUDE_{id}'


def escape_double_quotes(double_quotes):
    return double_quotes.replace('"', '\\"')


def unescape_double_quotes(escaped_double_quotes):
    return escaped_double_quotes.replace('\\"', '"')


def is_valid_include(include):
    is_valid_format = False
    for regex in RE_HOCON_INCLUDE:
        if re.findall(regex, include):
            is_valid_format = True
            break

    return is_valid_format


def get_include_key(include_str):
    """Use md5sum hash of the whole include statement string for a key.
    """
    return hashlib.md5(include_str.encode()).hexdigest()


def wrap_includes(hocon_str):
    """Convert `include` statement string into key = val format.
    Returns '{key} = "{double_quote_escaped_val}"'.
    """
    for regex in RE_HOCON_INCLUDE:
        for include in re.findall(regex, hocon_str):
            if '\\"' in include:
                continue

            logger.debug('Found include in HOCON: {include}'.format(include=include))

            hocon_str = hocon_str.replace(
                include,
                '{key} = "{val}"'.format(
                    key=HOCONSTRING_INCLUDE_KEY.format(id=get_include_key(include)),
                    val=escape_double_quotes(include),
                ),
            )
    return hocon_str


def unwrap_includes(key_val_str):
    """Convert '{key} = "{val}"" formatted string to the original `include` statement string.
    Args:
        key:
            HOCONSTRING_INCLUDE_KEY with `id` as md5sum hash of the original
            `include` statement string.
        val:
            Double-quote-escaped `include` statement string.
    """
    val = re.findall(RE_HOCONSTRING_INCLUDE_VALUE, key_val_str)
    if val:
        if len(val) > 1:
            raise ValueError(
                'Found multiple matches. Wrong include key=val format? {val}'.format(
                    val=val
                )
            )
        return unescape_double_quotes(val[0])


class HOCONString:
    def __init__(self, hocon_str):
        """Find an `include` statement (VALUE) in HOCON string and then convert it
        into a HOCONSTRING_INCLUDE_KEY="VALUE" pair in HOCON.

        Double-quotes will be escaped with double slashes.
        Then the VALUE is kept as it is as a value and can be recovered later when
        it is converted back to HOCON string.

        This workaround is to skip parsing `include` statements since there is no
        information about `classpath` at the parsing time and pyhocon will error out and
        will stop parsing.

        e.g. we don't know what's in `classpath` before the backend conf file is
        passed to Cromwell.
        """
        if not isinstance(hocon_str, str):
            raise ValueError('HOCONString() takes str type only.')

        self._hocon_str = wrap_includes(hocon_str)

    def __str__(self):
        return self.get_contents()

    @classmethod
    def from_dict(cls, d, include=''):
        """Create HOCONString from dict.

        Args:
            include:
                `include` statement to be added to the top of the HOCONString.
        """
        hocon = ConfigFactory.from_dict(d)
        hocon_str = HOCONConverter.to_hocon(hocon)

        if include:
            if not is_valid_include(include):
                raise ValueError(
                    'Wrong HOCON include format. {include}'.format(include=include)
                )
            hocon_str = NEW_LINE.join([include, hocon_str])

        return cls(hocon_str=hocon_str)

    def to_dict(self, with_include=True):
        """Convert HOCON string into dict.

        Args:
            with_include:
                If True then double-quote-escaped `include` statements will be kept as a plain string
                under key HOCONSTRING_INCLUDE_KEY.
                Otherwise, `include` statements will be excluded.
        """
        if with_include:
            hocon_str = self._hocon_str
        else:
            hocon_str = self.get_contents(with_include=False)

        c = ConfigFactory.parse_string(hocon_str)
        j = HOCONConverter.to_json(c)

        return json.loads(j)

    def merge(self, b, update=False):
        """Merge self with b and then returns a plain string of merged.
        Args:
            b:
                HOCONString, dict, str to be merged.
                b's `include` statement will always be ignored.
            update:
                If True then replace self with a merged one.
        Returns:
            String of merged HOCONs.
        """
        if isinstance(b, HOCONString):
            d = b.to_dict()
        elif isinstance(b, str):
            d = HOCONString(b).to_dict()
        elif isinstance(b, dict):
            d = b
        else:
            raise TypeError('Unsupported type {t}'.format(t=type(b)))

        self_d = self.to_dict()
        merge_dict(self_d, d)

        hocon = ConfigFactory.from_dict(self_d)

        hocon_str = HOCONConverter.to_hocon(hocon)
        if update:
            self._hocon_str = hocon_str

        return HOCONString(hocon_str).get_contents()

    def get_contents(self, with_include=True):
        """Check if `include` statement is stored as a plain string.
        If exists, converts it back to HOCON `include` statement.

        Args:
            with_include: (renamed/changed from without_include)
                If True then recover all includes statements from include key=val form
                (RE_HOCONSTRING_INCLUDE).
                Otherwise, excludes all `include` statements.
        """
        hocon_str = self._hocon_str

        for include_key_val in re.findall(RE_HOCONSTRING_INCLUDE, self._hocon_str):
            logger.debug(
                'Found include key in HOCONString: {include_key_val}'.format(
                    include_key_val=include_key_val
                )
            )
            if with_include:
                original_include_str = unwrap_includes(include_key_val)
                if original_include_str:
                    hocon_str = hocon_str.replace(include_key_val, original_include_str)
            else:
                hocon_str = hocon_str.replace(include_key_val, '')

        return hocon_str



import json
import os
from textwrap import dedent

from caper.singularity import Singularity

UBUNTU_18_04_3 = (
    'ubuntu@sha256:d1d454df0f579c6be4d8161d227462d69e163a8ff9d20a847533989cf0c94d90'
)
UBUNTU_18_04_3_LAST_HASH_TAR_GZ = (
    'sha256:6001e1789921cf851f6fb2e5fe05be70f482fe9c2286f66892fe5a3bc404569c.tar.gz'
)


def test_build_local_image(tmp_path):
    """Singularity class works with Singularity CLI.
    Make sure to have such CLI installed on your system.

    Make sure to unset env var SINGULARITY_CACHEDIR before running pytest.

    This test will build a local Singularity image, which is consist of
    multiple docker layer (hash) files (tar.gz).
    This test will check one of the hashes.
    """
    singularity = Singularity(
        singularity_image='docker://' + UBUNTU_18_04_3,
        singularity_cachedir=str(tmp_path),
    )
    singularity.build_local_image()

    hash_file = str(tmp_path / 'docker' / UBUNTU_18_04_3_LAST_HASH_TAR_GZ)
    assert os.path.exists(hash_file)


def test_find_bindpath(tmp_path):
    """Parse input JSON file to recursively get all the files defined in it.
    For found local abspaths, find common root directories for those.

    This is necessary for singularity to bind paths (similar to mounting directories
    in docker).

    Input JSON file has one TSV file and this file will be recursively visited by
    Singularilty.find_bindpath().
    """
    tsv = tmp_path / 'test.tsv'
    tsv_contents = dedent(
        """\
        file1\t/1/2/3/4.txt
        file2\t/1/5/6/7.txt
        file3\t/a/t/c.txt
    """
    )
    tsv.write_text(tsv_contents)

    inputs = tmp_path / 'inputs.json'
    inputs_dict = {
        'test.input_tsv': str(tsv),
        'test.input': '/a/b/c/d.txt',
        'test.input2': '/a/b/e.txt',
        'test.input3': '/f/g/h.txt',
        'test.input4': '/s/x/y/s/d/e/s/.txt',
    }
    inputs.write_text(json.dumps(inputs_dict, indent=4))

    # test with two different levels
    bindpaths_5 = Singularity.find_bindpath(str(inputs), 5).split(',')
    assert sorted(bindpaths_5) == sorted(
        [
            '/1/2/3',
            '/1/5/6',
            '/a/b',
            '/a/t',
            '/f/g',
            '/s/x/y/s',
            '/'.join(str(tmp_path).split('/')[:5]),
        ]
    )

    bindpaths_2 = Singularity.find_bindpath(str(inputs), 2).split(',')
    assert sorted(bindpaths_2) == sorted(
        ['/1', '/a', '/f', '/s', '/'.join(str(tmp_path).split('/')[:2])]
    )



import logging

from autouri import AutoURI

from .caper_base import CaperBase
from .caper_labels import CaperLabels
from .caper_wdl_parser import CaperWDLParser
from .caper_workflow_opts import CaperWorkflowOpts
from .cromwell import Cromwell
from .cromwell_rest_api import CromwellRestAPI, has_wildcard, is_valid_uuid
from .singularity import Singularity

logger = logging.getLogger(__name__)


class CaperClient(CaperBase):
    def __init__(
        self,
        local_loc_dir=None,
        gcp_loc_dir=None,
        aws_loc_dir=None,
        gcp_service_account_key_json=None,
        server_hostname=CromwellRestAPI.DEFAULT_HOSTNAME,
        server_port=CromwellRestAPI.DEFAULT_PORT,
        server_heartbeat=None,
    ):
        """Initializes for Caper's client functions.

        Args:
            server_hostname:
                Server hostname.
                Used only if heartbeat file is not available or timed out.
            server_port:
                Server port.
                Used only if heartbeat file is not available or timed out.
            server_heartbeat:
                ServerHeartbeat object in which a heartbeat file is defined.
                This object is to read hostname/port pair from it.
        """
        super().__init__(
            local_loc_dir=local_loc_dir,
            gcp_loc_dir=gcp_loc_dir,
            aws_loc_dir=aws_loc_dir,
            gcp_service_account_key_json=gcp_service_account_key_json,
        )

        if server_heartbeat:
            res = server_heartbeat.read()
            if res:
                server_hostname, server_port = res

        if not server_hostname or not server_port:
            raise ValueError(
                'Server hostname/port must be defined '
                'if server heartbeat is not available or timed out.'
            )

        self._cromwell_rest_api = CromwellRestAPI(server_hostname, server_port)

    def abort(self, wf_ids_or_labels):
        """Abort running/pending workflows on a Cromwell server.

        Args:
            wf_ids_or_labels:
                List of workflows IDs or string labels (Caper's string label)
                Wild cards (*, ?) are allowed.
        """
        workflow_ids, labels = self._split_workflow_ids_and_labels(wf_ids_or_labels)

        r = self._cromwell_rest_api.abort(workflow_ids, labels)
        logger.info('abort: {r}'.format(r=r))
        return r

    def unhold(self, wf_ids_or_labels):
        """Release hold of workflows on a Cromwell server.

        Args:
            wf_ids_or_labels:
                List of workflows IDs or string labels (Caper's string label)
                Wild cards (*, ?) are allowed.
        """
        workflow_ids, labels = self._split_workflow_ids_and_labels(wf_ids_or_labels)

        r = self._cromwell_rest_api.release_hold(workflow_ids, labels)
        logger.info('unhold: {r}'.format(r=r))
        return r

    def list(self, wf_ids_or_labels=None, exclude_subworkflow=True):
        """Retrieves list of running/pending workflows from a Cromwell server

        Args:
            wf_ids_or_labels:
                List of workflows IDs or string labels (Caper's string label)
                Wild cards (*, ?) are allowed.
            exclude_subworkflow:
                Exclude subworkflows
        Returns:
            List of workflows found. Each workflow object will be in a form of
            Cromwell's metadata JSON file but with limited amount of information.
            e.g. workflow ID, status, labels.
        """
        if wf_ids_or_labels:
            workflow_ids, labels = self._split_workflow_ids_and_labels(wf_ids_or_labels)
        else:
            workflow_ids, labels = ['*'], None

        return self._cromwell_rest_api.find(
            workflow_ids, labels, exclude_subworkflow=exclude_subworkflow
        )

    def metadata(self, wf_ids_or_labels, embed_subworkflow=False):
        """Retrieves metadata for workflows from a Cromwell server.

        Args:
            wf_ids_or_labels:
                List of workflows IDs or string labels (Caper's string label)
                Wild cards (*, ?) are allowed.
            embed_subworkflow:
                Recursively embed subworkflow's metadata JSON object
                in parent workflow's metadata JSON.
                This is to mimic behavior of Cromwell's run mode paramteter -m.
        Returns:
            List of metadata JSONs of matched worflows.
        """
        workflow_ids, labels = self._split_workflow_ids_and_labels(wf_ids_or_labels)

        return self._cromwell_rest_api.get_metadata(
            workflow_ids, labels, embed_subworkflow=embed_subworkflow
        )

    def _split_workflow_ids_and_labels(self, workflow_ids_or_labels):
        workflow_ids = []
        labels = []

        if workflow_ids_or_labels:
            for query in workflow_ids_or_labels:
                if is_valid_uuid(query):
                    workflow_ids.append(query)
                else:
                    labels.append((CaperLabels.KEY_CAPER_STR_LABEL, query))
                    if has_wildcard(query):
                        workflow_ids.append(query)

        return workflow_ids, labels


class CaperClientSubmit(CaperClient):
    def __init__(
        self,
        local_loc_dir=None,
        gcp_loc_dir=None,
        aws_loc_dir=None,
        gcp_service_account_key_json=None,
        server_hostname=CromwellRestAPI.DEFAULT_HOSTNAME,
        server_port=CromwellRestAPI.DEFAULT_PORT,
        server_heartbeat=None,
        womtool=Cromwell.DEFAULT_WOMTOOL,
        use_google_cloud_life_sciences=False,
        gcp_zones=None,
        slurm_partition=None,
        slurm_account=None,
        slurm_extra_param=None,
        sge_pe=None,
        sge_queue=None,
        sge_extra_param=None,
        pbs_queue=None,
        pbs_extra_param=None,
    ):
        """Submit subcommand needs much more parameters than other client subcommands.

        Args:
            womtool:
                Womtool JAR file.
            use_google_cloud_life_sciences:
                Use Google Cloud Life Sciences API.
                gcp_zones will be ignored since it's already configured with in
                server side backend.conf.
            gcp_zones:
                GCP zones. Used for gcp backend only.
            slurm_partition:
                SLURM partition if required to sbatch jobs.
            slurm_account:
                SLURM account if required to sbatch jobs.
            slurm_extra_param:
                SLURM extra parameter to be appended to sbatch command line.
            sge_pe:
                SGE parallel environment (required to run with multiple cpus).
            sge_queue:
                SGE queue.
            sge_extra_param:
                SGE extra parameter to be appended to qsub command line.
            pbs_queue:
                PBS queue.
            pbs_extra_param:
                PBS extra parameter to be appended to qsub command line.
        """
        super().__init__(
            local_loc_dir=local_loc_dir,
            gcp_loc_dir=gcp_loc_dir,
            aws_loc_dir=aws_loc_dir,
            gcp_service_account_key_json=gcp_service_account_key_json,
            server_hostname=server_hostname,
            server_port=server_port,
            server_heartbeat=server_heartbeat,
        )

        self._cromwell = Cromwell(womtool=womtool)

        self._caper_workflow_opts = CaperWorkflowOpts(
            use_google_cloud_life_sciences=use_google_cloud_life_sciences,
            gcp_zones=gcp_zones,
            slurm_partition=slurm_partition,
            slurm_account=slurm_account,
            slurm_extra_param=slurm_extra_param,
            sge_pe=sge_pe,
            sge_queue=sge_queue,
            sge_extra_param=sge_extra_param,
            pbs_queue=pbs_queue,
            pbs_extra_param=pbs_extra_param,
        )

        self._caper_labels = CaperLabels()

    def submit(
        self,
        wdl,
        backend=None,
        inputs=None,
        options=None,
        labels=None,
        imports=None,
        str_label=None,
        user=None,
        docker=None,
        singularity=None,
        singularity_cachedir=Singularity.DEFAULT_SINGULARITY_CACHEDIR,
        no_build_singularity=False,
        max_retries=CaperWorkflowOpts.DEFAULT_MAX_RETRIES,
        memory_retry_multiplier=CaperWorkflowOpts.DEFAULT_MEMORY_RETRY_MULTIPLIER,
        gcp_monitoring_script=CaperWorkflowOpts.DEFAULT_GCP_MONITORING_SCRIPT,
        ignore_womtool=False,
        no_deepcopy=False,
        hold=False,
        java_heap_womtool=Cromwell.DEFAULT_JAVA_HEAP_WOMTOOL,
        dry_run=False,
        work_dir=None,
    ):
        """Submit a workflow to Cromwell server.

        Args:
            wdl:
                WDL file.
            backend:
                Backend to run a workflow on.
                Choose among Caper's built-in or user's custom backends.
                (aws, gcp, Local, slurm, sge, pbs, ...).
                If not defined then server's default backend will be used.
            inputs:
                Input JSON file.
            options:
                Workflow options JSON file.
            labels:
                Labels JSON file.
            imports:
                imports ZIP file.
            str_label:
                Caper's string label for a workflow,
                which will be written to labels JSON file.
            user:
                Username. If not defined, find a username from system.
                This will be written to to Cromwell' labels JSON file and will not
                be used elsewhere.
            docker:
                Docker image to run a workflow on.
                This will add "docker" attribute to runtime {} section
                of all tasks in WDL.
                This will be overriden by existing "docker" attr defined in
                WDL's task's "runtime {} section.
            singularity:
                Singularity image to run a workflow on.
                To use this, do not define "docker" attribute in
                WDL's task's "runtime {} section.
            singularity_cachedir:
                Cache directory for local Singularity images.
                If there is a shell environment variable SINGULARITY_CACHEDIR
                define then this parameter will be ignored.
            no_build_singularity:
                Do not build local singularity image.
                However, a local singularity image will be eventually built on
                env var SINGULARITY_CACHEDIR.
                Therefore, use this flag if you have already built it.
            max_retries:
                Max retrial for a failed task. 0 or None means no trial.
            memory_retry_multiplier:
                Multiplier for the memory retry feature.
                See https://cromwell.readthedocs.io/en/develop/cromwell_features/RetryWithMoreMemory/
                for details.
            ignore_womtool:
                Disable Womtool validation for WDL/input JSON/imports.
            no_deepcopy:
                Disable recursive localization of files defined in input JSON.
                Input JSON file itself will still be localized.
            hold:
                Put a workflow on hold when submitted. This workflow will be on hold until
                it's released. See self.unhold() for details.
            java_heap_womtool:
                Java heap (java -Xmx) for Womtool.
            dry_run:
                Stop before running Java command line for Cromwell.
            work_dir:
                Local temporary directory to store all temporary files.
                Temporary files mean intermediate files used for running Cromwell.
                For example, workflow options file, imports zip file.
                Localized (recursively) data files defined in input JSON
                will NOT be stored here.
                They will be localized on self._local_loc_dir instead.
                If this is not defined, then cache directory self._local_loc_dir will be used.
        """
        wdl_file = AutoURI(wdl)
        if not wdl_file.exists:
            raise FileNotFoundError('WDL does not exists. {wdl}'.format(wdl=wdl))

        if str_label is None and inputs:
            str_label = AutoURI(inputs).basename_wo_ext

        if work_dir is None:
            work_dir = self.create_timestamped_work_dir(prefix=wdl_file.basename_wo_ext)

        wdl = wdl_file.localize_on(work_dir)

        if backend is None:
            backend = self._cromwell_rest_api.get_default_backend()

        if inputs:
            # inputs should be localized on corresponding
            # backend's localization directory.
            # check if such loc_dir is defined.
            if self.get_loc_dir(backend) is None:
                raise ValueError(
                    'loc_dir is not defined for your backend. {b}'.format(b=backend)
                )

            maybe_remote_file = self.localize_on_backend_if_modified(
                inputs, backend=backend, recursive=not no_deepcopy, make_md5_file=True
            )
            inputs = AutoURI(maybe_remote_file).localize_on(work_dir)

        options = self._caper_workflow_opts.create_file(
            directory=work_dir,
            wdl=wdl,
            backend=backend,
            inputs=inputs,
            custom_options=options,
            docker=docker,
            singularity=singularity,
            singularity_cachedir=singularity_cachedir,
            no_build_singularity=no_build_singularity,
            max_retries=max_retries,
            memory_retry_multiplier=memory_retry_multiplier,
            gcp_monitoring_script=gcp_monitoring_script,
        )

        labels = self._caper_labels.create_file(
            directory=work_dir,
            backend=backend,
            custom_labels=labels,
            str_label=str_label,
            user=user,
        )

        wdl_parser = CaperWDLParser(wdl)
        if imports:
            imports = AutoURI(imports).localize_on(work_dir)
        else:
            imports = wdl_parser.create_imports_file(work_dir)

        logger.debug(
            'submit params: wdl={wdl}, imports={imp}, inputs={inp}, '
            'options={opt}, labels={lbl}, hold={hold}'.format(
                wdl=wdl, imp=imports, inp=inputs, opt=options, lbl=labels, hold=hold
            )
        )

        if not ignore_womtool:
            self._cromwell.validate(
                wdl=wdl,
                inputs=inputs,
                imports=imports,
                java_heap_womtool=java_heap_womtool,
            )

        if dry_run:
            return

        r = self._cromwell_rest_api.submit(
            source=wdl,
            dependencies=imports,
            inputs=inputs,
            options=options,
            labels=labels,
            on_hold=hold,
        )
        logger.info('submit: {r}'.format(r=r))
        return r



from textwrap import dedent

from caper.caper_wdl_parser import CaperWDLParser

WDL_CONTENTS = dedent(
    """\
    version 1.0

    workflow test_wdl {
        meta {
            caper_docker: "ubuntu:latest"
            caper_singularity: "docker://ubuntu:latest"
        }
    }
"""
)


OLD_WDL_CONTENTS = dedent(
    """\
    #CAPER docker "ubuntu:latest"
    #CAPER singularity "docker://ubuntu:latest"

    workflow test_wdl {
    }
"""
)


def test_properties(tmp_path):
    """Test the following properties.
        - caper_docker
        - caper_singularity
    """
    main_wdl = tmp_path / 'main.wdl'
    main_wdl.write_text(WDL_CONTENTS)

    old_wdl = tmp_path / 'old_main.wdl'
    old_wdl.write_text(OLD_WDL_CONTENTS)

    # test reading from workflow.meta
    main = CaperWDLParser(str(main_wdl))
    assert main.caper_docker == 'ubuntu:latest'
    assert main.caper_singularity == 'docker://ubuntu:latest'

    # test reading from comments (old-style)
    old = CaperWDLParser(str(old_wdl))
    assert old.caper_docker == 'ubuntu:latest'
    assert old.caper_singularity == 'docker://ubuntu:latest'



"""dictTool: merge/split/flatten/unflatten dict

Author:
    Jin Lee (leepc12@gmail.com) at ENCODE-DCC
"""

import re
from collections import defaultdict

try:
    from collections.abc import MutableMapping
except AttributeError:
    from collections import MutableMapping


def merge_dict(a, b):
    """Merges b into a recursively. This mutates a and overwrites
    items in b on a for conflicts.

    Ref: https://stackoverflow.com/questions/7204805/dictionaries
    -of-dictionaries-merge/7205107#7205107
    """
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge_dict(a[key], b[key])
            elif a[key] == b[key]:
                pass
            else:
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a


def flatten_dict(d, reducer=None, parent_key=()):
    """Flattens dict into single-level-tuple-keyed dict with
        {(tuple of keys of parents and self): value}

    Args:
        reducer:
            Character to join keys in a tuple.
            If None, returns with key as a tuple.
    Returns:
        dict of {
            (key_lvl1, key_lvl2, key_lvl3, ...): value
        }
    """
    items = []
    for k, v in d.items():
        new_key = parent_key + (k if isinstance(k, tuple) else (k,))
        if isinstance(v, MutableMapping):
            items.extend(flatten_dict(v, parent_key=new_key).items())
        else:
            items.append((new_key, v))
    if reducer:
        return {reducer.join(k): v for k, v in type(d)(items).items()}
    else:
        return type(d)(items)


def recurse_dict_value(d, fnc):
    if isinstance(d, dict):
        for k, v in d.items():
            recurse_dict_value(v, fnc)

    elif isinstance(d, (list, tuple)):
        for v in d:
            recurse_dict_value(v, fnc)
    else:
        fnc(d)


def unflatten_dict(d_flat):
    """Unflattens single-level-tuple-keyed dict into dict
    """
    result = type(d_flat)()
    for k_tuple, v in d_flat.items():
        d_curr = result
        for i, k in enumerate(k_tuple):
            if i == len(k_tuple) - 1:
                d_curr[k] = v
            elif k not in d_curr:
                d_curr[k] = type(d_flat)()
            d_curr = d_curr[k]
    return result


def split_dict(d, rules=None):
    """Splits dict according to "rule"

    Returns:
        List of split dict

    Args:
        rule:
            A list of tuple (RULE_NAME: REGEX)

            If a key name in an JSON object matches with this REGEX
            then ALL objects with the same key will be separated from
            the original root JSON object while keeping their hierachy.
            RULE_NAME will be added to root of each new JSON object.

            For example, we have a JSON object like the following
            [
                {
                    "flagstat_qc": {
                        "rep1": {
                            "read1": 100,
                            "read2": 200
                        },
                        "rep2": {
                            "read1": 300,
                            "read2": 400
                        }
                    },
                    "etc": {
                        "samstat_qc": {
                            "rep1": {
                                "unmapped": 500,
                                "mapped": 600
                            },
                            "rep2": {
                                "unmapped": 700,
                                "mapped": 800
                            }
                        }
                    },
                    "idr_qc": {
                        "qc_test1" : 900
                    }
                }
            ]
            with "new_row_rule" = "replicate:^rep\\d+$", this JSON object
            will be splitted into three (original, rep1, rep2) JSON object.
            [
                # original
                {
                    "idr_qc": {
                        "qc_test1" : 900
                    }
                },
                # rep1
                {
                    "replicate": "rep1",
                    "flagstat_qc": {
                        "read1": 100,
                        "read2": 200
                    },
                    "etc": {
                        "samstat_qc": {
                            "unmapped": 500,
                            "mapped": 600
                        }
                    }
                },
                # rep2
                {
                    "replicate": "rep2",
                    "flagstat_qc": {
                        "read1": 300,
                        "read2": 400
                    },
                    "etc": {
                        "samstat_qc": {
                            "unmapped": 700,
                            "mapped": 800
                        }
                    }
                },
            ]
    """
    if rules is None:
        return [d]
    if isinstance(rules, tuple):
        rules = [rules]

    d_flat = flatten_dict(d)
    result = []
    keys_matched_regex = set()
    d_each_rule = defaultdict(type(d))
    for rule_name, rule_regex in rules:
        for k_tuple, v in d_flat.items():
            new_k_tuple = ()
            pattern_matched_k = None
            for k in k_tuple:
                if re.findall(rule_regex, k):
                    pattern_matched_k = (rule_name, k)
                else:
                    new_k_tuple += (k,)
            if pattern_matched_k is not None:
                d_each_rule[pattern_matched_k][new_k_tuple] = v
                keys_matched_regex.add(k_tuple)

    for (rule_name, k), d_each_matched in d_each_rule.items():
        d_ = unflatten_dict(d_each_matched)
        d_[rule_name] = k
        result.append(d_)

    d_others = type(d)()
    for k_tuple, v in d_flat.items():
        if k_tuple not in keys_matched_regex:
            d_others[k_tuple] = v
    if d_others:
        d_ = unflatten_dict(d_others)
        result = [d_] + result
    return result


def dict_to_dot_str(d, parent_key='digraph D', indent='', base_indent=''):
    """Dict will be converted into DOT like the followings:
        1) Value string will not be double-quotted in DOT.
            - make sure to escape double-quotes in a string with special characters
            (e.g. whitespace, # and ;)
        2) If "value" is None then "key" will be just added to DOT without "="

    dict:
        { "key1": "val1", "key2": "val2", "key3": { "key3_1": "val3_1", }... }

    dot:
        digraph D {
            key1 = val1;
            key2 = val2;
            key3 {
                key3_1 = val3_1;
                ...
            }
            ...
        }

    Example in a Croo output def JSON file:
        (note that strings for "label" are double-quote-escaped).

    dict:
        {
            "rankdir": "TD",
            "start": "[shape=Mdiamond]",
            "end": "[shape=Msquare]",
            "subgraph cluster_rep1": {
                "style": "filled",
                "color": "mistyrose",
                "label": "\"Replicate 1\""
            },
            "subgraph cluster_rep2": {
                "style": "filled",
                "color": "azure",
                "label": "\"Replicate 2\""
            },
            "a0 -> b0": null,
            "c0 -> d0": null
        }

    Such dict will be converted into a dot:

    dot:
        digraph D {
            rankDir = TD;
            start = [shape=Mdiamond];
            end = [shape=Msquare];
            subgraph cluster_rep1 {
                style = filled;
                color = mistyrose;
                label = "Replicate 1"
            };
            subgraph cluster_rep2 {
                style = filled;
                color = azure;
                label = "Replicate 2"
            };
            a0 -> b0;
            c0 -> d0;
        }
    """
    result = ''
    if d is None:
        return '{}{};\n'.format(base_indent, parent_key)
    elif isinstance(d, str):
        return '{}{} = {};\n'.format(base_indent, parent_key, d)
    elif isinstance(d, dict):
        result += base_indent + parent_key + ' {\n'
        for k, v in d.items():
            result += dict_to_dot_str(
                v, parent_key=k, indent=indent, base_indent=base_indent + indent
            )
        result += base_indent + '}\n'
    else:
        raise ValueError(
            'Unsupported data type: {} '
            '(only str and dict/JSON are allowed).'.format(type(d))
        )
    return result



import json
import logging
import os
import pwd
import re

from autouri import AutoURI

from .dict_tool import merge_dict

logger = logging.getLogger(__name__)


RE_ILLEGAL_STR_LABEL_CHRS = r'[\:\?\*]'
SUB_ILLEGAL_STR_LABEL_CHRS = '_'


class CaperLabels:
    KEY_CAPER_STR_LABEL = 'caper-str-label'
    KEY_CAPER_USER = 'caper-user'
    KEY_CAPER_BACKEND = 'caper-backend'
    BASENAME_LABELS = 'labels.json'

    def create_file(
        self,
        directory,
        backend=None,
        custom_labels=None,
        str_label=None,
        user=None,
        basename=BASENAME_LABELS,
    ):
        """Create labels JSON file.

        Args:
            directory:
                Directory to create a labels JSON file.
            backend:
                Backend
            custom_labels:
                User's labels file to be merged.
            str_label:
                Caper's string label.
                Wildcards ('*' and '?') and ':' are not allowed by default.
                These will be replaced with '_' by default.
            basename:
                Basename of labels file.
        """
        template = {}

        if custom_labels:
            s = AutoURI(custom_labels).read()
            merge_dict(template, json.loads(s))

        if backend:
            template[CaperLabels.KEY_CAPER_BACKEND] = backend

        if str_label:
            new_str_label = re.sub(
                RE_ILLEGAL_STR_LABEL_CHRS, SUB_ILLEGAL_STR_LABEL_CHRS, str_label
            )
            if str_label != new_str_label:
                logger.warning(
                    'Found illegal characters in str_label matching with {regex}. '
                    'Replaced with {sub}'.format(
                        regex=RE_ILLEGAL_STR_LABEL_CHRS, sub=SUB_ILLEGAL_STR_LABEL_CHRS
                    )
                )
            template[CaperLabels.KEY_CAPER_STR_LABEL] = new_str_label

        template[CaperLabels.KEY_CAPER_USER] = (
            user if user else pwd.getpwuid(os.getuid())[0]
        )

        labels_file = os.path.join(directory, basename)
        AutoURI(labels_file).write(json.dumps(template, indent=4))

        return labels_file



"""This does not cover all CLI parameters defined in caper/caper_args.py.
Google Cloud Platform is tested in test_cli_server_client.py.
However, other cloud (aws) and HPCs (slurm/sge/pbs) are not tested.

In this testing module, 'caper run' is tested with a local backend.

See test_cli_server_client.py for 'caper server/submit/...'.
We will use gcp (Google Cloud Platform) backend to test server-client
functions.
"""
import json
import os

import pytest
from autouri import GCSURI

from caper.cli import main as cli_main
from caper.cromwell_metadata import CromwellMetadata
from caper.wdl_parser import WDLParser

from .example_wdl import make_directory_with_wdls


def test_wrong_subcmd():
    cmd = ['wrong_subcmd']
    with pytest.raises(SystemExit):
        cli_main(cmd)


@pytest.mark.parametrize(
    'cmd',
    [
        ['--docker', '--singularity'],
        ['--docker', 'ubuntu:latest', '--singularity'],
        ['--docker', '--singularity', 'docker://ubuntu:latest'],
        ['--docker', 'ubuntu:latest', '--singularity', 'docker://ubuntu:latest'],
        ['--docker', '--soft-glob-output'],
        ['--docker', 'ubuntu:latest', '--soft-glob-output'],
    ],
)
def test_mutually_exclusive_params(tmp_path, cmd):
    make_directory_with_wdls(str(tmp_path))

    cmd = ['run', str(tmp_path / 'main.wdl')] + cmd
    with pytest.raises(ValueError):
        cli_main(cmd)


@pytest.mark.integration
def test_run(tmp_path, cromwell, womtool, debug_caper):
    """Will test most local parameters (run only) here.
    """
    make_directory_with_wdls(str(tmp_path))
    wdl = tmp_path / 'main.wdl'
    inputs = tmp_path / 'inputs.json'
    p = WDLParser(str(wdl))
    imports = p.zip_subworkflows(str(tmp_path / 'imports.zip'))

    cmd = ['run']
    cmd += [str(wdl)]
    cmd += ['--tmp-dir', str(tmp_path / 'tmp_dir')]
    # local (instead of correct Local with capital L) should work.
    cmd += ['--backend', 'local']
    cmd += ['--cromwell-stdout', str(tmp_path / 'cromwell_stdout.o')]
    cmd += ['--db', 'file']
    cmd += ['--db-timeout', '500000']
    cmd += ['--file-db', str(tmp_path / 'file_db_prefix')]
    cmd += ['--max-concurrent-tasks', '2']
    cmd += ['--max-concurrent-workflows', '2']
    cmd += ['--disable-call-caching']
    cmd += ['--soft-glob-output']
    cmd += ['--local-hash-strat', 'path']
    cmd += ['--local-out-dir', str(tmp_path / 'out_dir')]
    cmd += ['--inputs', str(inputs)]
    cmd += ['--imports', str(imports)]
    cmd += ['--ignore-womtool']
    cmd += ['--cromwell', cromwell]
    cmd += ['--womtool', womtool]
    cmd += ['--java-heap-womtool', '2G']
    cmd += ['--java-heap-run', '2G']
    cmd += ['--max-retries', '1']
    cmd += ['--metadata-output', str(tmp_path / 'metadata.json')]
    if debug_caper:
        cmd += ['--debug']

    cli_main(cmd)

    assert (tmp_path / 'tmp_dir').exists()
    assert (tmp_path / 'file_db_prefix.lobs').exists()
    assert (tmp_path / 'metadata.json').exists()
    assert (tmp_path / 'cromwell_stdout.o').exists()

    # test cleanup() on local storage
    cm = CromwellMetadata(str(tmp_path / 'metadata.json'))
    # check if metadata JSON and workflowRoot dir exists
    root_out_dir = cm.data['workflowRoot']
    assert os.path.exists(root_out_dir) and os.path.isdir(root_out_dir)

    # dry-run should not delete anything
    cm.cleanup(dry_run=True)
    assert os.path.exists(root_out_dir)

    cm.cleanup(dry_run=False)
    assert not os.path.exists(root_out_dir)


@pytest.mark.google_cloud
@pytest.mark.integration
def test_run_gcp_with_life_sciences_api(
    tmp_path,
    gcs_root,
    ci_prefix,
    cromwell,
    womtool,
    gcp_prj,
    gcp_service_account_key_json,
    debug_caper,
):
    """Test run with Google Cloud Life Sciences API
    """
    out_gcs_bucket = os.path.join(gcs_root, 'caper_out', ci_prefix)
    tmp_gcs_bucket = os.path.join(gcs_root, 'caper_tmp')

    # prepare WDLs and input JSON, imports to be submitted
    make_directory_with_wdls(str(tmp_path))
    wdl = tmp_path / 'main.wdl'
    inputs = tmp_path / 'inputs.json'
    metadata = tmp_path / 'metadata.json'

    cmd = ['run', str(wdl)]
    cmd += ['--inputs', str(inputs)]
    cmd += ['-m', str(metadata)]
    if gcp_service_account_key_json:
        cmd += ['--gcp-service-account-key-json', gcp_service_account_key_json]
    cmd += ['--use-google-cloud-life-sciences']
    cmd += ['--gcp-region', 'us-central1']
    # --gcp-zones should be ignored
    cmd += ['--gcp-zones', 'us-west1-a,us-west1-b']
    cmd += ['--gcp-prj', gcp_prj]
    cmd += ['--memory-retry-error-keys', 'Killed']
    cmd += ['--memory-retry-multiplier', '1.5']
    cmd += ['--tmp-dir', str(tmp_path / 'tmp_dir')]
    cmd += ['--backend', 'gcp']
    cmd += ['--gcp-out-dir', out_gcs_bucket]
    cmd += ['--gcp-loc-dir', tmp_gcs_bucket]
    cmd += ['--cromwell-stdout', str(tmp_path / 'cromwell_stdout.o')]
    # test with file type DB
    cmd += ['--db', 'file']
    cmd += ['--db-timeout', '500000']
    cmd += ['--file-db', str(tmp_path / 'file_db_prefix')]
    cmd += ['--max-concurrent-tasks', '2']
    cmd += ['--max-concurrent-workflows', '2']
    cmd += ['--disable-call-caching']
    cmd += ['--cromwell', cromwell]
    cmd += ['--womtool', womtool]
    cmd += ['--java-heap-run', '4G']
    cmd += ['--docker', 'ubuntu:latest']
    if debug_caper:
        cmd += ['--debug']
    print(' '.join(cmd))

    cli_main(cmd)
    m_dict = json.loads(metadata.read_text())

    assert m_dict['status'] == 'Succeeded'

    # test CromwellMetadata.gcp_monitor() here
    # since it's for gcp only and this function is one of the two
    # test functions ran on a gcp backend.
    # task main.t1 has sleep 10 so that monitoring_script has time to
    # write monitoring data to `monitoringLog` file
    cm = CromwellMetadata(m_dict)
    monitor_data = cm.gcp_monitor()
    for data in monitor_data:
        instance_cpu = data['instance']['cpu']
        instance_mem = data['instance']['mem']
        instance_disk = data['instance']['disk']
        assert instance_cpu >= 1
        assert instance_mem >= 1024 * 1024 * 1024
        assert instance_disk >= 10 * 1024 * 1024 * 1024

        max_cpu_percent = data['stats']['max']['cpu_pct']
        max_mem = data['stats']['max']['mem']
        max_disk = data['stats']['max']['disk']
        if max_cpu_percent or data['task_name'] == 'main.t1':
            assert max_cpu_percent <= 100.0
        if max_mem or data['task_name'] == 'main.t1':
            assert max_mem <= instance_mem
        if max_disk or data['task_name'] == 'main.t1':
            assert max_disk <= instance_disk

    # test cleanup on gcp backend (gs://)
    root_out_dir = cm.data['workflowRoot']

    # remote metadata JSON file on workflow's root output dir.
    remote_metadata_json_file = os.path.join(root_out_dir, 'metadata.json')
    assert GCSURI(remote_metadata_json_file).exists

    # dry-run should not delete anything
    cm.cleanup(dry_run=True)
    assert GCSURI(remote_metadata_json_file).exists

    cm.cleanup(dry_run=False)
    assert not GCSURI(remote_metadata_json_file).exists



import sys
import time

import pytest

from caper.caper_labels import CaperLabels
from caper.cromwell import Cromwell
from caper.cromwell_rest_api import CromwellRestAPI, has_wildcard, is_valid_uuid
from caper.wdl_parser import WDLParser

from .example_wdl import make_directory_with_wdls


@pytest.mark.parametrize(
    'test_input,expected',
    [
        ('asldkhjlkasdf289jisdl;sladkjasdflksd', False),
        ('cromwell-f9c26f2e-f550-4748-a650-5d0d4cab9f3a', False),
        ('f9c26f2e-f550-4748-a650-5d0d4c', False),
        ('f9c26f2e-f550-4748-a650-5d0d4cab9f3a', True),
        ('F9C26f2e-F550-4748-A650-5D0D4cab9f3a', False),
        ('f9c26f2e', False),
        ([], False),
        (tuple(), False),
        (None, False),
    ],
)
def test_is_valid_uuid(test_input, expected):
    assert is_valid_uuid(test_input) == expected


@pytest.mark.parametrize(
    'test_input,expected',
    [
        ('?????', True),
        (('lskadfj', 'sdkfjaslf'), False),
        ('*', True),
        ('?', True),
        (':', False),
        (('*', '?'), True),
        (('_', '-', 'asdfjkljklasdfjklasdf'), False),
        ([], False),
        (tuple(), False),
        (None, False),
    ],
)
def test_has_wildcard(test_input, expected):
    assert has_wildcard(test_input) == expected


def test_all(tmp_path, cromwell, womtool):
    """Test Cromwell.server() method, which returns a Thread object.
    """
    server_port = 8010
    fileobj_stdout = sys.stdout
    test_label = 'test_label'

    c = Cromwell(cromwell=cromwell, womtool=womtool)

    o_dir = tmp_path / 'output'
    o_dir.mkdir()

    labels_file = CaperLabels().create_file(
        directory=str(tmp_path), str_label=test_label
    )

    is_server_started = False

    def on_server_start():
        nonlocal is_server_started
        is_server_started = True

    workflow_id = None
    is_workflow_done = False

    def on_status_change(metadata):
        nonlocal workflow_id
        nonlocal is_workflow_done

        if metadata:
            if metadata['id'] == workflow_id:
                if metadata['status'] in ('Succeeded', 'Failed'):
                    is_workflow_done = True

    # also tests two callback functions
    try:
        th = c.server(
            server_port=server_port,
            embed_subworkflow=True,
            fileobj_stdout=fileobj_stdout,
            on_server_start=on_server_start,
            on_status_change=on_status_change,
            cwd=str(tmp_path),
        )
        assert th.status is None

        # wait until server is ready to take submissions
        t_start = time.time()
        while not is_server_started:
            time.sleep(1)
            if time.time() - t_start > 60:
                raise TimeoutError('Timed out waiting for Cromwell server spin-up.')

        # another way of checking server is started
        assert th.status

        # make WDLs and imports
        wdl = tmp_path / 'main.wdl'
        make_directory_with_wdls(str(tmp_path))
        # zip subworkflows for later use
        p = WDLParser(str(wdl))
        imports = p.zip_subworkflows(str(tmp_path / 'imports.zip'))

        cra = CromwellRestAPI(hostname='localhost', port=server_port)
        # no workflow
        assert not cra.find(workflow_ids=['*'])

        # put a hold on a workflow when submitting
        r = cra.submit(
            source=str(wdl),
            dependencies=imports,
            inputs=str(tmp_path / 'inputs.json'),
            labels=labels_file,
            on_hold=True,
        )
        workflow_id = r['id']
        time.sleep(10)
        # find by workflow ID
        workflow_by_id = cra.find(workflow_ids=[workflow_id])[0]
        # find by label
        workflow_by_label = cra.find(labels=[('caper-str-label', test_label)])[0]
        # find by workflow ID with wildcard *
        workflow_by_id_with_wildcard = cra.find(workflow_ids=[workflow_id[:-10] + '*'])[
            0
        ]
        # find by label with wildcard ?
        workflow_by_label_with_wildcard = cra.find(
            labels=[('caper-str-label', test_label[:-1] + '?')]
        )[0]

        assert workflow_by_label['id'] == workflow_id
        assert workflow_by_id['id'] == workflow_id
        assert workflow_by_id_with_wildcard['id'] == workflow_id
        assert workflow_by_label_with_wildcard['id'] == workflow_id
        assert workflow_by_id['status'] == 'On Hold'

        cra.release_hold([workflow_id])
        time.sleep(3)

        assert cra.get_label(workflow_id, 'caper-str-label') == test_label
        assert cra.get_labels(workflow_id)['caper-str-label'] == test_label

        # abort it
        assert cra.find([workflow_id])[0]['status'] in ('Submitted', 'On Hold')
        cra.abort([workflow_id])
        time.sleep(5)
        assert cra.find([workflow_id])[0]['status'] == 'Aborted'

        # submit another workflow
        r = cra.submit(
            source=str(wdl),
            dependencies=imports,
            inputs=str(tmp_path / 'inputs.json'),
            on_hold=False,
        )
        is_workflow_done = False
        workflow_id = r['id']
        time.sleep(5)

        t_start = time.time()
        while not is_workflow_done:
            time.sleep(1)
            print('polling: ', workflow_id, is_workflow_done)
            if time.time() - t_start > 120:
                raise TimeoutError('Timed out waiting for workflow being done.')

        metadata = cra.get_metadata([workflow_id], embed_subworkflow=True)[0]
        metadata_wo_sub = cra.get_metadata([workflow_id], embed_subworkflow=False)[0]

        assert 'subWorkflowMetadata' not in metadata_wo_sub['calls']['main.sub'][0]
        subworkflow = metadata['calls']['main.sub'][0]
        assert 'subWorkflowMetadata' in subworkflow
        assert (
            'subWorkflowMetadata'
            in subworkflow['subWorkflowMetadata']['calls']['sub.sub_sub'][0]
        )

        # check server's properties before closing it
        assert cra.get_default_backend() == 'Local'
        assert cra.get_backends()['supportedBackends'] == ['Local']

    finally:
        th.stop()
        th.join()



import os
import time

import pytest

from caper.nb_subproc_thread import NBSubprocThread

SH_CONTENTS = """#!/bin/bash

echoerr() { echo "$@" 1>&2; }

NUM=$1
if [ -z "$NUM" ]
then
  NUM=10
fi

echo $NUM

# NUM kitties (1 kitty per sec)
for i in $(seq 1 $NUM)
do
  echo "hello kitty $i-1. (STDOUT)"
  sleep 0.25
  echoerr "hello kitty $i-1. (STDERR)"
  sleep 0.25
  echoerr "hello kitty $i-2. (STDERR)"
  sleep 0.25
  echo "hello kitty $i-2. (STDOUT)"
  sleep 0.25
done

exit 10
"""


def on_stdout(stdout):
    print('captured stdout:', stdout)
    assert stdout.endswith('\n')


def on_stderr(stderr):
    print('captured stderr:', stderr)
    assert stderr.endswith('\n')


def on_poll():
    print('polling')


def on_finish():
    return 'done'


def test_nb_subproc_thread(tmp_path):
    sh = tmp_path / 'test.sh'
    sh.write_text(SH_CONTENTS)

    th = NBSubprocThread(
        args=['bash', str(sh)],
        on_poll=on_poll,
        on_stdout=on_stdout,
        on_stderr=on_stderr,
        on_finish=on_finish,
        poll_interval=0.1,
    )
    assert th.returnvalue is None
    assert not th.is_alive()
    th.start()
    assert th.is_alive()
    # rc is None while running
    assert th.returncode is None
    th.join()
    assert th.returncode == 10
    assert th.returnvalue == 'done'


def test_nb_subproc_thread_stopped(tmp_path):
    sh = tmp_path / 'test.sh'
    sh.write_text(SH_CONTENTS)

    th = NBSubprocThread(args=['bash', str(sh)], on_stdout=on_stdout)
    th.start()
    time.sleep(2)
    assert th.is_alive()
    th.stop(wait=True)
    assert not th.is_alive()
    # rc should be is None if terminated
    assert th.returncode is not None
    # subprocess is terminated until it reaches kitty 4 (4 sec > 2 sec).
    assert 'hello kitty 4' not in th.stderr


def test_nb_subproc_thread_nonzero_rc():
    for rc in range(10):
        th = NBSubprocThread(
            args=['bash', '-c', 'exit {rc}'.format(rc=rc)], on_stderr=on_stderr
        )
        th.start()
        th.join()
        assert th.returncode == rc


@pytest.mark.parametrize('test_app,expected_rc', [('cat', 1), ('ls', 2), ('java', 1)])
def test_nb_subproc_thread_nonzero_rc_for_real_apps(test_app, expected_rc):
    test_str = 'asdfasf-10190212-zxcv'
    if os.path.exists(test_str):
        raise ValueError('Test string should not be an existing file.')

    th = NBSubprocThread(
        args=[test_app, test_str], on_stdout=on_stdout, on_stderr=on_stderr
    )
    th.start()
    th.join()
    assert th.returncode == expected_rc
    assert test_str in th.stderr
    assert th.stdout == ''



import logging
import os
import re
import shutil
from tempfile import TemporaryDirectory

from autouri import HTTPURL, AbsPath, AutoURI
from WDL import parse_document

logger = logging.getLogger(__name__)


class WDLParser:
    RE_WDL_IMPORT = r'^\s*import\s+[\"\'](.+)[\"\']\s*'
    RECURSION_DEPTH_LIMIT = 20
    BASENAME_IMPORTS = 'imports.zip'

    def __init__(self, wdl):
        """Wraps miniwdl's parse_document().
        """
        u = AutoURI(wdl)
        if not u.exists:
            raise FileNotFoundError('WDL does not exist: wdl={wdl}'.format(wdl=wdl))
        self._wdl = wdl
        self._wdl_contents = AutoURI(wdl).read()
        try:
            self._wdl_doc = parse_document(self._wdl_contents)
        except Exception:
            logger.error('Failed to parse WDL with miniwdl.')
            self._wdl_doc = None

    @property
    def contents(self):
        return self._wdl_contents

    @property
    def workflow_meta(self):
        if self._wdl_doc:
            return self._wdl_doc.workflow.meta

    @property
    def workflow_parameter_meta(self):
        if self._wdl_doc:
            return self._wdl_doc.workflow.parameter_meta

    @property
    def imports(self):
        """Miniwdl (0.3.7) has a bug for URL imports.
        Keep using reg-ex to find imports until it's fixed.
        Returns:
            List of URIs of imported subworkflows.
        """
        try:
            return [i.uri for i in self._wdl_doc.imports]
        except Exception:
            pass
        return self._find_val_of_matched_lines(WDLParser.RE_WDL_IMPORT)

    def zip_subworkflows(self, zip_file):
        """Recursively find/zip imported subworkflow WDLs
        This will zip sub-WDLs with relative paths only.
        i.e. URIs are ignored.
        For this (main) workflow, any URI is allowed.
        However, only subworkflows with relative path will be zipped
        since there is no way to make directory structure to zip them.
        Returns:
            Zipped imports file.
            None if no subworkflows recursively found in WDL.
        """
        with TemporaryDirectory() as tmp_d:
            # localize WDL first. If it's already local
            # then will use its original path without loc.
            wdl = AutoURI(self._wdl).localize_on(tmp_d)
            # keep directory structure as they imported
            num_sub_wf_packed = self.__recurse_zip_subworkflows(
                root_zip_dir=tmp_d, root_wdl_dir=AutoURI(wdl).dirname
            )
            if num_sub_wf_packed:
                shutil.make_archive(AutoURI(zip_file).uri_wo_ext, 'zip', tmp_d)
                return zip_file

    def create_imports_file(self, directory, basename=BASENAME_IMPORTS):
        """Wrapper for zip_subworkflows.
        This creates an imports zip file with basename on directory.
        """
        zip_file = os.path.join(directory, basename)
        if self.zip_subworkflows(zip_file):
            return zip_file

    def _find_val_of_matched_lines(self, regex, no_strip=False):
        """Find value of the first line matching regex.
        Args:
            regex:
                Regular expression. This should have only one ().
            no_strip:
                Do not strip result strings.
        Returns:
            Value of the first line matching regex.
        """
        res = []
        for line in self.contents.split('\n'):
            r = re.findall(regex, line)
            if len(r) > 0:
                res.append(r[0] if no_strip else r[0].strip())
        return res

    def __recurse_zip_subworkflows(
        self, root_zip_dir, root_wdl_dir, imported_as_url=False, depth=0
    ):
        """Recurse imported sub-WDLs in main-WDL.
        Unlike Cromwell, Womtool does not take imports.zip while validating WDLs.
        All sub-WDLs should be in a correct directory structure relative to the
        root WDL.
        For Womtool, we should make a temporary directory and unpack imports.zip there and
        need to make a copy of root WDL on it. Then run Womtool to validate them.
        This function is to make such imports.zip.
        Sub-WDLs imported as relative path simply inherit parent's directory.
        Sub-WDLs imported as URL does not inherit parent's directory but root
        WDL's directory.
        Sub-WDLs imported as absolute path are not allowed. This can work with "caper run"
        but not with "caper submit" (or Cromwell submit).
        Args:
            depth: Recursion depth
        Returns:
            Total number of subworkflows:
                Sub WDL files "recursively" localized on "root_zip_dir".
        """
        if depth > WDLParser.RECURSION_DEPTH_LIMIT:
            raise ValueError(
                'Reached recursion depth limit while zipping subworkflows recursively. '
                'Possible cyclic import or self-refencing in WDLs? wdl={wdl}'.format(
                    wdl=self._wdl
                )
            )

        if imported_as_url:
            main_wdl_dir = root_wdl_dir
        else:
            main_wdl_dir = AbsPath(self._wdl).dirname

        num_sub_wf_packed = 0
        for sub_rel_to_parent in self.imports:
            sub_wdl_file = AutoURI(sub_rel_to_parent)

            if isinstance(sub_wdl_file, HTTPURL):
                sub_abs = sub_wdl_file.uri
                imported_as_url_sub = True
            elif isinstance(sub_wdl_file, AbsPath):
                raise ValueError(
                    'For sub WDL zipping, absolute path is not allowed for sub WDL. '
                    'main={main}, sub={sub}'.format(
                        main=self._wdl, sub=sub_rel_to_parent
                    )
                )
            else:
                sub_abs = os.path.realpath(
                    os.path.join(main_wdl_dir, sub_rel_to_parent)
                )
                if not AbsPath(sub_abs).exists:
                    raise FileNotFoundError(
                        'Sub WDL does not exist. Did you import main WDL '
                        'as a URL but sub WDL references a local file? '
                        'main={main}, sub={sub}, imported_as_url={i}'.format(
                            main=self._wdl, sub=sub_rel_to_parent, i=imported_as_url
                        )
                    )
                if not sub_abs.startswith(root_wdl_dir):
                    raise ValueError(
                        'Sub WDL exists but it is out of root WDL directory. '
                        'Too many "../" in your sub WDL? '
                        'Or main WDL is imported as an URL but sub WDL '
                        'has "../"? '
                        'main={main}, sub={sub}, imported_as_url={i}'.format(
                            main=self._wdl, sub=sub_rel_to_parent, i=imported_as_url
                        )
                    )

                # make a copy on zip_dir
                rel_path = os.path.relpath(sub_abs, root_wdl_dir)
                cp_dest = os.path.join(root_zip_dir, rel_path)

                AbsPath(sub_abs).cp(cp_dest)
                num_sub_wf_packed += 1
                imported_as_url_sub = False

            num_sub_wf_packed += WDLParser(sub_abs).__recurse_zip_subworkflows(
                root_zip_dir=root_zip_dir,
                root_wdl_dir=root_wdl_dir,
                imported_as_url=imported_as_url_sub,
                depth=depth + 1,
            )
        return num_sub_wf_packed



"""
There are three WDL 1.0 files to test subworkflow zipping.
main.wdl (imports sub/sub.wdl)
    sub/
        sub.wdl (imports sub/sub_sub.wdl)
        sub/
            sub_sub.wdl
There is another trivial WDL 1.0 file with empty workflow.
"""

import os
import shutil

from caper.wdl_parser import WDLParser

from .example_wdl import (
    MAIN_WDL,
    MAIN_WDL_META_DICT,
    MAIN_WDL_PARAMETER_META_DICT,
    make_directory_with_wdls,
)


def test_properties(tmp_path):
    """Test the following properties.
        - contents
        - workflow_meta
        - workflow_parameter_meta
        - imports
    """
    wdl = tmp_path / 'main.wdl'
    wdl.write_text(MAIN_WDL)

    wp = WDLParser(str(wdl))
    assert wp.contents == MAIN_WDL
    assert wp.workflow_meta == MAIN_WDL_META_DICT
    assert wp.workflow_parameter_meta == MAIN_WDL_PARAMETER_META_DICT
    assert wp.imports == ['sub/sub.wdl']


def test_zip_subworkflows(tmp_path):
    """This actually tests create_imports_file since
    create_imports_file's merely a wrapper for zip_subworkflows.
    """
    # make tmp directory to store WDLs
    make_directory_with_wdls(str(tmp_path))

    # we now have all WDL files
    # main.wdl, sub/sub.wdl, sub/sub/sub_sub.wdl

    main_wdl = tmp_path / 'main.wdl'
    sub_sub_wdl = tmp_path / 'sub' / 'sub' / 'sub_sub.wdl'

    main = WDLParser(str(main_wdl))

    # simple WDL without any imports
    simple = WDLParser(str(sub_sub_wdl))

    # make working directory
    d = tmp_path / 'imports'
    d.mkdir(parents=True)

    simple_zip_file = simple.create_imports_file(str(d), 'test_trivial_imports.zip')
    assert simple_zip_file is None

    main_zip_file = main.create_imports_file(str(d), 'test_imports.zip')
    assert os.path.basename(main_zip_file) == 'test_imports.zip'

    shutil.unpack_archive(main_zip_file, extract_dir=str(d))
    assert os.path.exists(str(d / 'sub' / 'sub.wdl'))
    assert os.path.exists(str(d / 'sub' / 'sub' / 'sub_sub.wdl'))



#!/usr/bin/env python3
"""
"""
import pytest

from caper.cromwell import Cromwell


def pytest_addoption(parser):
    parser.addoption(
        '--ci-prefix', default='default_ci_prefix', help='Prefix for CI test.'
    )
    parser.addoption(
        '--gcs-root',
        default='gs://encode-test-caper',
        help='GCS root path for CI test. '
        'This GCS bucket must be publicly accessible '
        '(read access for everyone is enough for testing).',
    )
    parser.addoption(
        '--cromwell',
        default=Cromwell.DEFAULT_CROMWELL,
        help='URI for Cromwell JAR. Local path is recommended.',
    )
    parser.addoption(
        '--womtool',
        default=Cromwell.DEFAULT_WOMTOOL,
        help='URI for Womtool JAR. Local path is recommended.',
    )
    parser.addoption(
        '--gcp-prj', default='encode-dcc-1016', help='Project on Google Cloud Platform.'
    )
    parser.addoption(
        '--gcp-service-account-key-json', help='JSON key file for GCP service account.'
    )
    parser.addoption(
        '--debug-caper', action='store_true', help='Debug-level logging for CLI tests.'
    )


@pytest.fixture(scope='session')
def ci_prefix(request):
    return request.config.getoption('--ci-prefix').rstrip('/')


@pytest.fixture(scope='session')
def gcs_root(request):
    """GCS root to generate test GCS URIs on.
    """
    return request.config.getoption('--gcs-root').rstrip('/')


@pytest.fixture(scope='session')
def cromwell(request):
    return request.config.getoption('--cromwell')


@pytest.fixture(scope='session')
def womtool(request):
    return request.config.getoption('--womtool')


@pytest.fixture(scope='session')
def gcp_prj(request):
    return request.config.getoption('--gcp-prj')


@pytest.fixture(scope='session')
def gcp_service_account_key_json(request):
    return request.config.getoption('--gcp-service-account-key-json')


@pytest.fixture(scope='session')
def debug_caper(request):
    return request.config.getoption('--debug-caper')


@pytest.fixture(scope='session')
def gcp_res_analysis_metadata():
    return 'gs://caper-data/gcp_resource_analysis/out/atac/e5eab444-cb6c-414a-a090-2c12417be542/metadata.json'



import logging
import os

from autouri import AbsPath, AutoURI

from .caper_backend_conf import CaperBackendConf
from .caper_base import CaperBase
from .caper_labels import CaperLabels
from .caper_workflow_opts import CaperWorkflowOpts
from .cromwell import Cromwell
from .cromwell_backend import (
    CromwellBackendAWS,
    CromwellBackendBase,
    CromwellBackendCommon,
    CromwellBackendDatabase,
    CromwellBackendGCP,
    CromwellBackendLocal,
)
from .cromwell_metadata import CromwellMetadata
from .cromwell_rest_api import CromwellRestAPI
from .singularity import Singularity
from .wdl_parser import WDLParser

logger = logging.getLogger(__name__)


class WomtoolValidationFailedException(Exception):
    pass


class CaperRunner(CaperBase):
    ENV_GOOGLE_CLOUD_PROJECT = 'GOOGLE_CLOUD_PROJECT'
    DEFAULT_FILE_DB_PREFIX = 'default_caper_file_db'
    SERVER_TMP_DIR_PREFIX = '.caper_server'

    def __init__(
        self,
        default_backend,
        local_loc_dir=None,
        local_out_dir=None,
        gcp_loc_dir=None,
        aws_loc_dir=None,
        cromwell=Cromwell.DEFAULT_CROMWELL,
        womtool=Cromwell.DEFAULT_WOMTOOL,
        disable_call_caching=False,
        max_concurrent_workflows=CromwellBackendCommon.DEFAULT_MAX_CONCURRENT_WORKFLOWS,
        memory_retry_error_keys=CromwellBackendCommon.DEFAULT_MEMORY_RETRY_ERROR_KEYS,
        max_concurrent_tasks=CromwellBackendBase.DEFAULT_CONCURRENT_JOB_LIMIT,
        soft_glob_output=False,
        local_hash_strat=CromwellBackendLocal.DEFAULT_LOCAL_HASH_STRAT,
        db=CromwellBackendDatabase.DEFAULT_DB,
        db_timeout=CromwellBackendDatabase.DEFAULT_DB_TIMEOUT_MS,
        mysql_db_ip=CromwellBackendDatabase.DEFAULT_MYSQL_DB_IP,
        mysql_db_port=CromwellBackendDatabase.DEFAULT_MYSQL_DB_PORT,
        mysql_db_user=CromwellBackendDatabase.DEFAULT_MYSQL_DB_USER,
        mysql_db_password=CromwellBackendDatabase.DEFAULT_MYSQL_DB_PASSWORD,
        mysql_db_name=CromwellBackendDatabase.DEFAULT_MYSQL_DB_NAME,
        postgresql_db_ip=CromwellBackendDatabase.DEFAULT_POSTGRESQL_DB_IP,
        postgresql_db_port=CromwellBackendDatabase.DEFAULT_POSTGRESQL_DB_PORT,
        postgresql_db_user=CromwellBackendDatabase.DEFAULT_POSTGRESQL_DB_USER,
        postgresql_db_password=CromwellBackendDatabase.DEFAULT_POSTGRESQL_DB_PASSWORD,
        postgresql_db_name=CromwellBackendDatabase.DEFAULT_POSTGRESQL_DB_NAME,
        file_db=None,
        gcp_prj=None,
        gcp_out_dir=None,
        gcp_call_caching_dup_strat=CromwellBackendGCP.DEFAULT_CALL_CACHING_DUP_STRAT,
        gcp_service_account_key_json=None,
        use_google_cloud_life_sciences=False,
        gcp_region=CromwellBackendGCP.DEFAULT_REGION,
        aws_batch_arn=None,
        aws_region=None,
        aws_out_dir=None,
        aws_call_caching_dup_strat=CromwellBackendAWS.DEFAULT_CALL_CACHING_DUP_STRAT,
        gcp_zones=None,
        slurm_partition=None,
        slurm_account=None,
        slurm_extra_param=None,
        sge_pe=None,
        sge_queue=None,
        sge_extra_param=None,
        pbs_queue=None,
        pbs_extra_param=None,
    ):
        """See docstring of base class for other arguments.

        Args:
            default_backend:
                Default backend.
            cromwell:
                Cromwell JAR URI.
            womtool:
                Womtool JAR URI.
            disable_call_caching:
            max_concurrent_workflows:
            memory_retry_error_keys
            max_concurrent_tasks:
            soft_glob_output:
            local_hash_strat:
            db:
            db_timeout:
            mysql_db_ip:
            mysql_db_port:
            mysql_db_user:
            mysql_db_password:
            mysql_db_name:
            postgresql_db_ip:
            postgresql_db_port:
            postgresql_db_user:
            postgresql_db_password:
            postgresql_db_name:
            file_db:
            gcp_prj:
            gcp_call_caching_dup_strat:
            gcp_service_account_key_json:
                This will be added to environment variable
                GOOGLE_APPLICATION_CREDENTIALS
                If not match with existing key then error out.
            use_google_cloud_life_sciences:
                Use Google Cloud Life Sciences API instead of Genomics API
                which has beed deprecated.
            gcp_region:
                Region for Google Cloud Life Sciences API.
                Ignored if not use_google_cloud_life_sciences.
            gcp_out_dir:
            aws_batch_arn:
            aws_region:
            aws_out_dir:
            aws_call_caching_dup_strat:
            gcp_zones:
                For this and all below arguments,
                see details in CaperWorkflowOpts.__init__.
            slurm_partition:
                SLURM partition if required to sbatch jobs.
            slurm_account:
                SLURM account if required to sbatch jobs.
            slurm_extra_param:
                SLURM extra parameter to be appended to sbatch command line.
            sge_pe:
                SGE parallel environment (required to run with multiple cpus).
            sge_queue:
                SGE queue.
            sge_extra_param:
                SGE extra parameter to be appended to qsub command line.
            pbs_queue:
                PBS queue.
            pbs_extra_param:
                PBS extra parameter to be appended to qsub command line.
        """
        super().__init__(
            local_loc_dir=local_loc_dir,
            gcp_loc_dir=gcp_loc_dir,
            aws_loc_dir=aws_loc_dir,
            gcp_service_account_key_json=gcp_service_account_key_json,
        )
        self._set_env_gcp_prj(gcp_prj)

        self._cromwell = Cromwell(cromwell=cromwell, womtool=womtool)

        if local_out_dir is None:
            local_out_dir = os.getcwd()

        self._caper_backend_conf = CaperBackendConf(
            default_backend=default_backend,
            local_out_dir=local_out_dir,
            disable_call_caching=disable_call_caching,
            max_concurrent_workflows=max_concurrent_workflows,
            max_concurrent_tasks=max_concurrent_tasks,
            soft_glob_output=soft_glob_output,
            local_hash_strat=local_hash_strat,
            db=db,
            db_timeout=db_timeout,
            file_db=file_db,
            mysql_db_ip=mysql_db_ip,
            mysql_db_port=mysql_db_port,
            mysql_db_user=mysql_db_user,
            mysql_db_password=mysql_db_password,
            mysql_db_name=mysql_db_name,
            postgresql_db_ip=postgresql_db_ip,
            postgresql_db_port=postgresql_db_port,
            postgresql_db_user=postgresql_db_user,
            postgresql_db_password=postgresql_db_password,
            postgresql_db_name=postgresql_db_name,
            gcp_prj=gcp_prj,
            gcp_out_dir=gcp_out_dir,
            memory_retry_error_keys=memory_retry_error_keys,
            gcp_call_caching_dup_strat=gcp_call_caching_dup_strat,
            gcp_service_account_key_json=gcp_service_account_key_json,
            use_google_cloud_life_sciences=use_google_cloud_life_sciences,
            gcp_region=gcp_region,
            aws_batch_arn=aws_batch_arn,
            aws_region=aws_region,
            aws_out_dir=aws_out_dir,
            aws_call_caching_dup_strat=aws_call_caching_dup_strat,
            gcp_zones=gcp_zones,
            slurm_partition=slurm_partition,
            slurm_account=slurm_account,
            slurm_extra_param=slurm_extra_param,
            sge_pe=sge_pe,
            sge_queue=sge_queue,
            sge_extra_param=sge_extra_param,
            pbs_queue=pbs_queue,
            pbs_extra_param=pbs_extra_param,
        )

        self._caper_workflow_opts = CaperWorkflowOpts(
            use_google_cloud_life_sciences=use_google_cloud_life_sciences,
            gcp_zones=gcp_zones,
            slurm_partition=slurm_partition,
            slurm_account=slurm_account,
            slurm_extra_param=slurm_extra_param,
            sge_pe=sge_pe,
            sge_queue=sge_queue,
            sge_extra_param=sge_extra_param,
            pbs_queue=pbs_queue,
            pbs_extra_param=pbs_extra_param,
        )

        self._caper_labels = CaperLabels()

    def _set_env_gcp_prj(self, gcp_prj=None, env_name=ENV_GOOGLE_CLOUD_PROJECT):
        """Initalizes environment for authentication (storage).
        Args:
            gcp_prj:
                Environment variable GOOGLE_CLOUD_PROJECT will be updated with
                this.
        """
        if gcp_prj:
            if env_name in os.environ:
                prj = os.environ[env_name]
                if prj != gcp_prj:
                    logger.warning(
                        'Env var {env} does not match with '
                        'gcp_prj {prj}.'.format(env=env_name, prj=gcp_prj)
                    )
            logger.debug(
                'Adding {prj} to env var {env}'.format(prj=gcp_prj, env=env_name)
            )
            os.environ[env_name] = gcp_prj

    def run(
        self,
        backend,
        wdl,
        inputs=None,
        options=None,
        labels=None,
        imports=None,
        metadata_output=None,
        str_label=None,
        user=None,
        docker=None,
        singularity=None,
        singularity_cachedir=Singularity.DEFAULT_SINGULARITY_CACHEDIR,
        no_build_singularity=False,
        custom_backend_conf=None,
        max_retries=CaperWorkflowOpts.DEFAULT_MAX_RETRIES,
        memory_retry_multiplier=CaperWorkflowOpts.DEFAULT_MEMORY_RETRY_MULTIPLIER,
        gcp_monitoring_script=CaperWorkflowOpts.DEFAULT_GCP_MONITORING_SCRIPT,
        ignore_womtool=False,
        no_deepcopy=False,
        fileobj_stdout=None,
        fileobj_troubleshoot=None,
        work_dir=None,
        java_heap_run=Cromwell.DEFAULT_JAVA_HEAP_CROMWELL_RUN,
        java_heap_womtool=Cromwell.DEFAULT_JAVA_HEAP_WOMTOOL,
        dry_run=False,
    ):
        """Run a workflow using Cromwell run mode.

        Args:
            backend:
                Choose among Caper's built-in backends.
                (aws, gcp, Local, slurm, sge, pbs).
                Or use a backend defined in your custom backend config file
                (above "backend_conf" file).
            wdl:
                WDL file.
            inputs:
                Input JSON file. Cromwell's parameter -i.
            options:
                Workflow options JSON file. Cromwell's parameter -o.
            labels:
                Labels JSON file. Cromwell's parameter -l.
            imports:
                imports ZIP file. Cromwell's parameter -p.
            metadata_output:
                Output metadata file path. Metadata JSON file will be written to
                this path. Caper also automatiacally generates it on each workflow's
                root directory.  Cromwell's parameter -m.
            str_label:
                Caper's string label, which will be written
                to labels JSON object.
            user:
                Username. If not defined, find username from system.
            docker:
                Docker image to run a workflow on.
                This will be overriden by "docker" attr defined in
                WDL's task's "runtime {} section.

                If this is None:
                    Docker will not be used for this workflow.
                If this is an emtpy string (working like a flag):
                    Docker will be used. Caper will try to find docker image
                    in WDL (from a comment "#CAPER docker" or
                    from workflow.meta.caper_docker).
            singularity:
                Singularity image to run a workflow on.
                This will be overriden by "singularity" attr defined in
                WDL's task's "runtime {} section.

                If this is None:
                    Singularity will not be used for this workflow.
                If this is an emtpy string:
                    Singularity will be used. Caper will try to find Singularity image
                    in WDL (from a comment "#CAPER singularity" or
                    from workflow.meta.caper_singularlity).
            singularity_cachedir:
                Cache directory for local Singularity images.
                If there is a shell environment variable SINGULARITY_CACHEDIR
                define then this parameter will be ignored.
            no_build_singularity:
                Do not build local singularity image.
                However, a local singularity image will be eventually built on
                env var SINGULARITY_CACHEDIR.
                Therefore, use this flag if you have already built it.
            custom_backend_conf:
                Backend config file (HOCON) to override Caper's
                auto-generated backend config.
            max_retries:
                Number of retrial for a failed task in a workflow.
                This applies to every task in a workflow.
                0 means no retrial. "attemps" attribute in a task's metadata
                increments from 1 as it is retried. attempts==2 means first retrial.
            memory_retry_multiplier:
                Multiplier for the memory retry feature.
                See https://cromwell.readthedocs.io/en/develop/cromwell_features/RetryWithMoreMemory/
                for details.
            ignore_womtool:
                Disable Womtool validation for WDL/input JSON/imports.
            no_deepcopy:
                Disable recursive localization of files defined in input JSON.
                Input JSON file itself will still be localized.
            fileobj_stdout:
                File-like object to write Cromwell's STDOUT.
            fileobj_troubleshoot:
                File-like object to write auto-troubleshooting after workflow is done.
            work_dir:
                Local temporary directory to store all temporary files.
                Temporary files mean intermediate files used for running Cromwell.
                For example, backend config file, workflow options file.
                Localized (recursively) data files defined in input JSON
                will NOT be stored here.
                They will be localized on self._local_loc_dir instead.
                If this is not defined, then cache directory self._local_loc_dir will be used.
                However, Cromwell Java process itself will run on CWD instead of this directory.
            java_heap_run:
                Java heap (java -Xmx) for Cromwell server mode.
            java_heap_womtool:
                Java heap (java -Xmx) for Womtool.
            dry_run:
                Stop before running Java command line for Cromwell.
        Returns:
            metadata_file:
                URI of metadata JSON file.
        """
        if not AutoURI(wdl).exists:
            raise FileNotFoundError('WDL does not exists. {wdl}'.format(wdl=wdl))

        if str_label is None and inputs:
            str_label = AutoURI(inputs).basename_wo_ext

        if work_dir is None:
            work_dir = self.create_timestamped_work_dir(
                prefix=AutoURI(wdl).basename_wo_ext
            )

        logger.info('Localizing files on work_dir. {d}'.format(d=work_dir))

        if inputs:
            maybe_remote_file = self.localize_on_backend_if_modified(
                inputs, backend=backend, recursive=not no_deepcopy, make_md5_file=True
            )
            inputs = AutoURI(maybe_remote_file).localize_on(work_dir)

        if imports:
            imports = AutoURI(imports).localize_on(work_dir)
        elif not AbsPath(wdl).exists:
            # auto-zip sub WDLs only if main WDL is remote
            imports = WDLParser(wdl).create_imports_file(work_dir)

        # localize WDL to be passed to Cromwell Java
        wdl = AutoURI(wdl).localize_on(work_dir)

        if metadata_output:
            if not AbsPath(metadata_output).is_valid:
                raise ValueError(
                    'metadata_output is not a valid local abspath. {m}'.format(
                        m=metadata_output
                    )
                )
        else:
            metadata_output = os.path.join(
                work_dir, CromwellMetadata.DEFAULT_METADATA_BASENAME
            )

        backend_conf = self._caper_backend_conf.create_file(
            directory=work_dir, backend=backend, custom_backend_conf=custom_backend_conf
        )

        options = self._caper_workflow_opts.create_file(
            directory=work_dir,
            wdl=wdl,
            inputs=inputs,
            custom_options=options,
            docker=docker,
            singularity=singularity,
            singularity_cachedir=singularity_cachedir,
            backend=backend,
            max_retries=max_retries,
            memory_retry_multiplier=memory_retry_multiplier,
            gcp_monitoring_script=gcp_monitoring_script,
        )

        labels = self._caper_labels.create_file(
            directory=work_dir,
            backend=backend,
            custom_labels=labels,
            str_label=str_label,
            user=user,
        )

        if not ignore_womtool:
            self._cromwell.validate(wdl=wdl, inputs=inputs, imports=imports)

        logger.info(
            'launching run: wdl={w}, inputs={i}, backend_conf={b}'.format(
                w=wdl, i=inputs, b=backend_conf
            )
        )
        th = self._cromwell.run(
            wdl=wdl,
            backend_conf=backend_conf,
            inputs=inputs,
            options=options,
            imports=imports,
            labels=labels,
            metadata=metadata_output,
            fileobj_stdout=fileobj_stdout,
            fileobj_troubleshoot=fileobj_troubleshoot,
            dry_run=dry_run,
        )
        return th

    def server(
        self,
        default_backend,
        server_port=CromwellRestAPI.DEFAULT_PORT,
        server_hostname=None,
        server_heartbeat=None,
        custom_backend_conf=None,
        fileobj_stdout=None,
        embed_subworkflow=False,
        java_heap_server=Cromwell.DEFAULT_JAVA_HEAP_CROMWELL_SERVER,
        auto_write_metadata=True,
        work_dir=None,
        dry_run=False,
    ):
        """Run a Cromwell server.
            default_backend:
                Default backend. If backend is not specified for a submitted workflow
                then default backend will be used.
                Choose among Caper's built-in backends.
                (aws, gcp, Local, slurm, sge, pbs).
                Or use a backend defined in your custom backend config file
                (above "backend_conf" file).
            server_heartbeat:
                Server heartbeat to write hostname/port of a server.
            server_port:
                Server port to run Cromwell server.
                Make sure to use different port for multiple Cromwell servers on the same
                machine.
            server_hostname:
                Server hostname. If not defined then socket.gethostname() will be used.
                If server_heartbeat is given, then this hostname will be written to
                the server heartbeat file defined in server_heartbeat.
            custom_backend_conf:
                Backend config file (HOCON) to override Caper's auto-generated backend config.
            fileobj_stdout:
                File-like object to write Cromwell's STDOUT.
            embed_subworkflow:
                Caper stores/updates metadata.JSON file on
                each workflow's root directory whenever there is status change
                of workflow (or its tasks).
                This flag ensures that any subworkflow's metadata JSON will be
                embedded in main (this) workflow's metadata JSON.
                This is to mimic behavior of Cromwell run mode's -m parameter.
            java_heap_server:
                Java heap (java -Xmx) for Cromwell server mode.
            auto_write_metadata:
                Automatic retrieval/writing of metadata.json upon workflow/task's status change.
            work_dir:
                Local temporary directory to store all temporary files.
                Temporary files mean intermediate files used for running Cromwell.
                For example, auto-generated backend config file and workflow options file.
                If this is not defined, then cache directory self._local_loc_dir with a timestamp
                will be used.
                However, Cromwell Java process itself will run on CWD instead of this directory.
            dry_run:
                Stop before running Java command line for Cromwell.
        """
        if work_dir is None:
            work_dir = self.create_timestamped_work_dir(
                prefix=CaperRunner.SERVER_TMP_DIR_PREFIX
            )

        backend_conf = self._caper_backend_conf.create_file(
            directory=work_dir,
            backend=default_backend,
            custom_backend_conf=custom_backend_conf,
        )
        logger.info('launching server: backend_conf={b}'.format(b=backend_conf))

        th = self._cromwell.server(
            backend_conf=backend_conf,
            server_port=server_port,
            server_hostname=server_hostname,
            server_heartbeat=server_heartbeat,
            fileobj_stdout=fileobj_stdout,
            embed_subworkflow=embed_subworkflow,
            java_heap_cromwell_server=java_heap_server,
            auto_write_metadata=auto_write_metadata,
            dry_run=dry_run,
        )
        return th



import os
import sys

from autouri import AutoURI

from caper.cromwell import Cromwell
from caper.cromwell_metadata import CromwellMetadata

from .example_wdl import make_directory_with_failing_wdls, make_directory_with_wdls


def test_on_successful_workflow(tmp_path, cromwell, womtool):
    fileobj_stdout = sys.stdout

    make_directory_with_wdls(str(tmp_path / 'successful'))

    # Run Cromwell to get metadata JSON
    c = Cromwell(cromwell=cromwell, womtool=womtool)
    th = c.run(
        wdl=str(tmp_path / 'successful' / 'main.wdl'),
        inputs=str(tmp_path / 'successful' / 'inputs.json'),
        fileobj_stdout=fileobj_stdout,
        cwd=str(tmp_path / 'successful'),
    )
    th.join()
    metadata = th.returnvalue
    assert metadata

    cm = CromwellMetadata(metadata)
    # test all properties
    assert cm.data == metadata
    assert cm.metadata == metadata
    assert CromwellMetadata(metadata).data == metadata
    assert cm.workflow_id == metadata['id']
    assert cm.workflow_status == metadata['status']
    # no failures for successful workflow's metadata
    assert cm.failures is None
    assert cm.calls == metadata['calls']
    assert sorted([call_name for call_name, _, _ in cm.recursed_calls]) == sorted(
        ['main.t1', 'sub.t2', 'sub_sub.t3']
    )

    # test recurse_calls(): test with a simple function
    def fnc(call_name, call, parent_call_names):
        assert call_name in ('main.t1', 'sub.t2', 'sub_sub.t3')
        assert call['executionStatus'] == 'Done'
        if call_name == 'main.t1':
            assert not parent_call_names
        elif call_name == 'sub.t2':
            assert parent_call_names == ('main.sub',)
        elif call_name == 'sub_sub.t3':
            assert parent_call_names == ('main.sub', 'sub.sub_sub')
        else:
            raise ValueError('Wrong call_name: {name}'.format(name=call_name))

    list(cm.recurse_calls(fnc))

    # test write_on_workflow_root()
    m_file_on_root = os.path.join(cm.metadata['workflowRoot'], 'metadata.json')
    u = AutoURI(m_file_on_root)
    u.rm()
    assert not u.exists

    cm.write_on_workflow_root()
    assert os.path.exists(m_file_on_root)
    assert CromwellMetadata(m_file_on_root).metadata == cm.metadata


def test_on_failed_workflow(tmp_path, cromwell, womtool):
    fileobj_stdout = sys.stdout

    make_directory_with_failing_wdls(str(tmp_path / 'failed'))

    # Run Cromwell to get metadata JSON
    # designed to fail in a subworkflow
    c = Cromwell(cromwell=cromwell, womtool=womtool)
    th = c.run(
        wdl=str(tmp_path / 'failed' / 'main.wdl'),
        inputs=str(tmp_path / 'failed' / 'inputs.json'),
        fileobj_stdout=fileobj_stdout,
        cwd=str(tmp_path / 'failed'),
    )
    th.join()

    # check failed
    assert th.returncode
    metadata = th.returnvalue
    assert metadata
    cm = CromwellMetadata(metadata)

    assert cm.failures == metadata['failures']
    assert cm.calls == metadata['calls']

    # test troubleshoot()
    report = cm.troubleshoot()
    assert '* Found failures JSON object' in report
    assert 'NAME=sub.t2_failing' in report
    assert 'INTENTED_ERROR: command not found' in report



"""Test is based on a metadata JSON file generated from
running atac-seq-pipeline v1.8.0 with the following input JSON.
gs://encode-pipeline-test-samples/encode-atac-seq-pipeline/ENCSR356KRQ_subsampled_caper.json
"""

import pytest

from caper.resource_analysis import LinearResourceAnalysis, ResourceAnalysis


def test_resource_analysis_abstract_class(gcp_res_analysis_metadata):
    with pytest.raises(TypeError):
        # abstract base-class
        ResourceAnalysis()


def test_resource_analysis_analyze_task(gcp_res_analysis_metadata):
    analysis = LinearResourceAnalysis()
    analysis.collect_resource_data([gcp_res_analysis_metadata])

    result_align1 = analysis.analyze_task(
        'atac.align',
        in_file_vars=['fastqs_R1'],
        reduce_in_file_vars=None,
        target_resources=['stats.max.mem', 'stats.mean.cpu_pct'],
    )
    assert result_align1['x'] == {'fastqs_R1': [15643136, 18963919]}
    assert 'stats.mean.cpu_pct' in result_align1['y']
    assert 'stats.max.mem' in result_align1['y']
    assert 'stats.max.disk' not in result_align1['y']
    assert list(result_align1['y'].keys()) == list(result_align1['coeffs'].keys())
    assert result_align1['coeffs']['stats.mean.cpu_pct'][0][0] == pytest.approx(
        1.6844513715565233e-06
    )
    assert result_align1['coeffs']['stats.mean.cpu_pct'][1] == pytest.approx(
        42.28561239506905
    )
    assert result_align1['coeffs']['stats.max.mem'][0][0] == pytest.approx(
        48.91222341236991
    )
    assert result_align1['coeffs']['stats.max.mem'][1] == pytest.approx(
        124314029.09791338
    )

    result_align2 = analysis.analyze_task(
        'atac.align', in_file_vars=['fastqs_R2'], reduce_in_file_vars=sum
    )
    assert result_align2['x'] == {'sum(fastqs_R2)': [16495088, 20184668]}
    assert 'stats.mean.cpu_pct' not in result_align2['y']
    assert 'stats.max.mem' in result_align2['y']
    assert 'stats.max.disk' in result_align2['y']
    assert list(result_align2['y'].keys()) == list(result_align2['coeffs'].keys())

    result_align_star = analysis.analyze_task('atac.align*', reduce_in_file_vars=max)
    assert result_align_star['x'] == {
        'max(chrsz,fastqs_R1,fastqs_R2,idx_tar,tmp_fastqs)': [
            32138224,
            39148587,
            3749246230,
            3749246230,
        ]
    }


def test_resource_analysis_analyze(gcp_res_analysis_metadata):
    """Test method analyze() which analyze all tasks defined in in_file_vars.
    """
    analysis = LinearResourceAnalysis()
    analysis.collect_resource_data([gcp_res_analysis_metadata])

    result = analysis.analyze(
        in_file_vars={
            'atac.align*': ['fastqs_R1', 'fastqs_R2'],
            'atac.filter*': ['bam'],
        }
    )
    assert len(result) == 2
    assert result['atac.align*']['x'] == {
        'sum(fastqs_R1,fastqs_R2)': [32138224, 39148587, 32138224, 39148587]
    }
    assert result['atac.filter*']['x'] == {
        'sum(bam)': [61315022, 76789196, 61315022, 76789196]
    }

    result_all = analysis.analyze()
    # 38 tasks in total
    assert len(result_all) == 38



import copy
import json
import logging
import os

from autouri import GCSURI, AutoURI

from .caper_wdl_parser import CaperWDLParser
from .cromwell_backend import BACKEND_AWS, BACKEND_GCP
from .dict_tool import merge_dict
from .singularity import Singularity

logger = logging.getLogger(__name__)


class CaperWorkflowOpts:
    DEFAULT_RUNTIME_ATTRIBUTES = 'default_runtime_attributes'
    BASENAME_WORKFLOW_OPTS_JSON = 'workflow_opts.json'
    DEFAULT_MAX_RETRIES = 1
    DEFAULT_MEMORY_RETRY_MULTIPLIER = 1.2
    DEFAULT_GCP_MONITORING_SCRIPT = (
        'gs://caper-data/scripts/resource_monitor/resource_monitor.sh'
    )

    def __init__(
        self,
        use_google_cloud_life_sciences=False,
        gcp_zones=None,
        slurm_partition=None,
        slurm_account=None,
        slurm_extra_param=None,
        sge_pe=None,
        sge_queue=None,
        sge_extra_param=None,
        pbs_queue=None,
        pbs_extra_param=None,
    ):
        """Template for a workflows options JSON file.
        All parameters are optional.

        Args:
            use_google_cloud_life_sciences:
                Use Google Cloud Life Sciences API instead of Genomics API
                which has beed deprecated.
                If this flag is on gcp_zones is ignored.
            gcp_zones:
                For gcp backend only.
                List of GCP zones to run workflows on.
            slurm_partition:
                For slurm backend only.
                SLURM partition to submit tasks to.
                Caper will submit tasks with "sbatch --partition".
            slurm_account:
                For slurm backend only.
                SLURM account to submit tasks to.
                Caper will submit tasks with "sbatch --account".
            slurm_extra_param:
                For slurm backend only.
                Extra parameters for SLURM.
                This will be appended to "sbatch" command line.
            sge_pe:
                For sge backend only.
                Name of parallel environment (PE) of SGE cluster.
                If it does not exist ask your admin to add one.
            sge_queue:
                For sge backend only.
                SGE queue to submit tasks to.
            sge_extra_param:
                For sge backend only.
                Extra parameters for SGE.
                This will be appended to "qsub" command line.
            pbs_queue:
                For pbs backend only.
                PBS queue to submit tasks to.
            pbs_extra_param:
                For pbs backend only.
                Extra parameters for PBS.
                This will be appended to "qsub" command line.
        """
        self._template = {CaperWorkflowOpts.DEFAULT_RUNTIME_ATTRIBUTES: dict()}
        dra = self._template[CaperWorkflowOpts.DEFAULT_RUNTIME_ATTRIBUTES]

        if gcp_zones and not use_google_cloud_life_sciences:
            dra['zones'] = ' '.join(gcp_zones)

        if slurm_partition:
            dra['slurm_partition'] = slurm_partition
        if slurm_account:
            dra['slurm_account'] = slurm_account
        if slurm_extra_param:
            dra['slurm_extra_param'] = slurm_extra_param

        if sge_pe:
            dra['sge_pe'] = sge_pe
        if sge_queue:
            dra['sge_queue'] = sge_queue
        if sge_extra_param:
            dra['sge_extra_param'] = sge_extra_param

        if pbs_queue:
            dra['pbs_queue'] = pbs_queue
        if pbs_extra_param:
            dra['pbs_extra_param'] = pbs_extra_param

    def create_file(
        self,
        directory,
        wdl,
        backend=None,
        inputs=None,
        custom_options=None,
        docker=None,
        singularity=None,
        singularity_cachedir=None,
        no_build_singularity=False,
        max_retries=DEFAULT_MAX_RETRIES,
        memory_retry_multiplier=DEFAULT_MEMORY_RETRY_MULTIPLIER,
        gcp_monitoring_script=DEFAULT_GCP_MONITORING_SCRIPT,
        basename=BASENAME_WORKFLOW_OPTS_JSON,
    ):
        """Creates Cromwell's workflow options JSON file.
        Workflow options JSON file sets default values for attributes
        defined in runtime {} section of WDL's task.
        For example, docker attribute can be defined here instead of directory
        defining in task's runtime { docker: "" }.

        Args:
            directory:
                Directory to make workflow options JSON file.
            wdl:
                WDL file.
            backend:
                Backend to run a workflow on. If not defined, server's default or
                runner's Local backend will be used.
            inputs:
                Input JSON file to define input files/parameters for WDL.
                This will be overriden by environment variable SINGULARITY_BINDPATH.
                For Singularity, it is required to find SINGULARITY_BINDPATH,
                which is a comma-separated list of common root directories
                for all files defined in input JSON.
                Unlike Docker, Singularity binds directories instead of mounting them.
                Therefore, Caper will try to find an optimal SINGULARITY_BINDPATH
                by looking at all files paths and find common parent directories for them.
            custom_options:
                User's custom workflow options JSON file.
                This will be merged at the end of this function.
                Therefore, users can override on Caper's auto-generated
                workflow options JSON file.
            docker:
                Docker image to run a workflow on.
            singularity:
                Singularity image to run a workflow on.
            singularity_cachedir:
                Singularity cache directory to build local images on.
                This will be overriden by environment variable SINGULARITY_CACHEDIR.
            no_build_singularity:
                Caper run "singularity exec IMAGE" to build a local Singularity image
                before submitting/running a workflow.
                With this flag on, Caper does not pre-build a local Singularity container.
                Therefore, Singularity container will be built inside each task,
                which will result in multiple redundant local image building.
                Also, trying to build on the same Singularity image file can
                lead to corruption of the image file.
            max_retries:
                Maximum number of retirals for each task. 1 means 1 retrial.
            memory_retry_multiplier:
                Multiplier for the memory retry feature.
                See https://cromwell.readthedocs.io/en/develop/cromwell_features/RetryWithMoreMemory/
                for details.
            gcp_monitoring_script:
                Monitoring script for GCP backend only.
                Useful to monitor resources on an instance.
            basename:
                Basename for a temporary workflow options JSON file.
        """
        if singularity and docker:
            raise ValueError('Cannot use both Singularity and Docker.')

        template = copy.deepcopy(self._template)
        dra = template[CaperWorkflowOpts.DEFAULT_RUNTIME_ATTRIBUTES]

        if backend:
            template['backend'] = backend

        wdl_parser = CaperWDLParser(wdl)
        if docker == '' or backend in (BACKEND_GCP, BACKEND_AWS) and not docker:
            # find "caper-docker" from WDL's workflow.meta
            # or "#CAPER docker" from comments
            docker = wdl_parser.caper_docker
            if docker:
                logger.info(
                    'Docker image found in WDL\'s metadata. wdl={wdl}, d={d}'.format(
                        wdl=wdl, d=docker
                    )
                )
            else:
                logger.warning(
                    "Docker image not found in WDL's metadata, which means that "
                    "docker is not defined either as comment (#CAPER docker) or "
                    "in workflow's meta section (under key caper_docker) in WDL. "
                    "If your WDL already has docker defined "
                    "in each task's runtime "
                    "then it should be okay. wdl={wdl}".format(wdl=wdl)
                )
        if docker:
            dra['docker'] = docker

        if singularity == '':
            if backend in (BACKEND_GCP, BACKEND_AWS):
                raise ValueError(
                    'Singularity cannot be used for cloud backend (e.g. aws, gcp).'
                )

            singularity = wdl_parser.caper_singularity
            if singularity:
                logger.info(
                    'Singularity image found in WDL\'s metadata. wdl={wdl}, s={s}'.format(
                        wdl=wdl, s=singularity
                    )
                )
            else:
                raise ValueError(
                    'Singularity image not found in WDL. wdl={wdl}'.format(wdl=wdl)
                )
        if singularity:
            dra['singularity'] = singularity
            if singularity_cachedir:
                dra['singularity_cachedir'] = singularity_cachedir

            s = Singularity(singularity, singularity_cachedir)
            if inputs:
                dra['singularity_bindpath'] = s.find_bindpath(inputs)
            if not no_build_singularity:
                s.build_local_image()

        if max_retries is not None:
            dra['maxRetries'] = max_retries
        # Cromwell's bug in memory-retry feature.
        # Disabled until it's fixed on Cromwell's side.
        # if memory_retry_multiplier is not None:
        #     template['memory_retry_multiplier'] = memory_retry_multiplier

        if gcp_monitoring_script and backend == BACKEND_GCP:
            if not GCSURI(gcp_monitoring_script).is_valid:
                raise ValueError(
                    'gcp_monitoring_script is not a valid URI. {uri}'.format(
                        uri=gcp_monitoring_script
                    )
                )
            template['monitoring_script'] = gcp_monitoring_script

        if custom_options:
            s = AutoURI(custom_options).read()
            d = json.loads(s)
            merge_dict(template, d)

        final_options_file = os.path.join(directory, basename)
        AutoURI(final_options_file).write(json.dumps(template, indent=4) + '\n')

        return final_options_file



import socket
import time

import pytest

from caper.server_heartbeat import ServerHeartbeat, ServerHeartbeatTimeoutError


def test_server_heartbeat(tmp_path):
    """All methods will be tested here.
    This willl test 3 things:
        - can read from file
        - can get hostname of this machine
        - can ignore old file (> heartbeat_timeout of 5 sec)
    """
    hb_file = tmp_path / 'hb_file'

    hb = ServerHeartbeat(heartbeat_file=str(hb_file), heartbeat_timeout=5000)

    # before starting write thread
    # it should return None
    assert hb.read() is None

    try:
        hb.start(port=9999)

        time.sleep(1)
        assert hb.read() == (socket.gethostname(), 9999)
    finally:
        hb.stop()

    # after timeout
    time.sleep(5)
    assert hb.read() is None

    with pytest.raises(ServerHeartbeatTimeoutError):
        hb.read(raise_timeout=True)



import logging
import os
from subprocess import check_call

from autouri import AbsPath, AutoURI, URIBase
from autouri.loc_aux import recurse_json

logger = logging.getLogger(__name__)


class Singularity:
    DEFAULT_SINGULARITY_CACHEDIR = '~/.caper/singularity_cachedir'
    DEFAULT_COMMON_ROOT_SEARCH_LEVEL = 5

    def __init__(
        self, singularity_image, singularity_cachedir=DEFAULT_SINGULARITY_CACHEDIR
    ):
        """This class is based on Singularity CLI.
        """
        self._singularity_image = singularity_image
        self._singularity_cachedir = singularity_cachedir

    def build_local_image(self):
        """Build local image for Singularity on SINGULARITY_CACHEDIR.

        Args.
            singularity_cachedir:
                Cache directory for local Singularity images.
                If there is a shell environment variable SINGULARITY_CACHEDIR
                define then this parameter will be ignored.
        """
        singularity_cachedir = os.path.abspath(
            os.path.expanduser(self._singularity_cachedir)
        )
        os.makedirs(singularity_cachedir, exist_ok=True)

        cmd = [
            'singularity',
            'exec',
            self._singularity_image,
            'echo',
            'Built local singularity image for {img}'.format(
                img=self._singularity_image
            ),
        ]
        logger.info(
            'Building local singularity image for {img}'.format(
                img=self._singularity_image
            )
        )

        env = os.environ.copy()
        if self._singularity_cachedir and 'SINGULARITY_CACHEDIR' not in env:
            env['SINGULARITY_CACHEDIR'] = self._singularity_cachedir

        return check_call(cmd, env=env)

    @staticmethod
    def find_bindpath(
        json_file, common_root_search_level=DEFAULT_COMMON_ROOT_SEARCH_LEVEL
    ):
        """Recursively find paths to be bound for singularity.
        Find common roots for all files in an input JSON file.
        This function will recursively visit all values in input JSON and
        also JSON, TSV, CSV files in the input JSON itself.

        This function visit all files in input JSON.
        Files with some extensions (defined by Autouri's URIBase.LOC_RECURSE_EXT_AND_FNC)
        are recursively visited.

        Add all (but not too high level<4) parent directories
        to all_dirnames. start from original
        For example, we have /a/b/c/d/e/f/g/h with common_root_search_level = 5
            add all the followings:
            /a/b/c/d/e/f/g/h (org)
            /a/b/c/d/e/f/g
            /a/b/c/d/e/f
            /a/b/c/d/e
            /a/b/c/d (minimum level = COMMON_ROOT_SEARCH_LEVEL-1)

        Args:
            json_file:
                Input JSON file which have local paths in it.
                Non-path values will be just ignored.
            common_root_search_level:
                See above description.
        """
        json_contents = AutoURI(json_file).read()
        all_dirnames = []

        def find_dirname(s):
            u = AbsPath(s)
            if u.is_valid:
                for ext, recurse_fnc_for_ext in URIBase.LOC_RECURSE_EXT_AND_FNC.items():
                    if u.ext == ext:
                        _, _ = recurse_fnc_for_ext(u.read(), find_dirname)
                # file can be a soft-link
                # singularity will want to have access to both soft-link and real one
                # so add dirnames of both soft-link and realpath
                all_dirnames.append(u.dirname)
                all_dirnames.append(os.path.dirname(os.path.realpath(u.uri)))
            return None, False

        _, _ = recurse_json(json_contents, find_dirname)

        all_dnames_incl_parents = set(all_dirnames)
        for d in all_dirnames:
            dir_arr = d.split(os.sep)
            for i, _ in enumerate(dir_arr[common_root_search_level:]):
                d_child = os.sep.join(dir_arr[: i + common_root_search_level])
                all_dnames_incl_parents.add(d_child)

        bindpaths = set()
        # remove overlapping directories
        for i, d1 in enumerate(sorted(all_dnames_incl_parents, reverse=True)):
            overlap_found = False
            for j, d2 in enumerate(sorted(all_dnames_incl_parents, reverse=True)):
                if i >= j:
                    continue
                if d1.startswith(d2):
                    overlap_found = True
                    break
            if not overlap_found:
                bindpaths.add(d1)

        return ','.join(bindpaths)



"""There are lots of UserDict-based classesi n caper/cromwell_backend.py
In this test, only the followings classes with public methods
will be tested.
    - CromwellBackendBase

"""
from caper.cromwell_backend import CromwellBackendBase


def test_cromwell_backend_base_backend():
    """Test a property backend's getter, setter
    """
    bb1 = CromwellBackendBase('test1')
    backend_dict = {'a': 1, 'b': '2'}

    bb1.backend = backend_dict
    assert bb1.backend == backend_dict


def test_cromwell_backend_base_merge_backend():
    bb1 = CromwellBackendBase('test1')
    bb1.backend = {'a': 1, 'b': '2'}
    backend_dict = {'c': 3.0, 'd': '4.0'}

    bb1.merge_backend(backend_dict)
    assert bb1.backend == {'a': 1, 'b': '2', 'c': 3.0, 'd': '4.0'}


def test_cromwell_backend_base_backend_config():
    bb1 = CromwellBackendBase('test1')
    bb1.backend = {'config': {'root': 'test/folder'}}
    assert bb1.backend_config == {'root': 'test/folder'}


def test_cromwell_backend_base_backend_config_dra():
    bb1 = CromwellBackendBase('test1')
    bb1.backend = {
        'config': {
            'root': 'test/folder',
            'default-runtime-attributes': {'docker': 'ubuntu:latest'},
        }
    }
    assert bb1.default_runtime_attributes == {'docker': 'ubuntu:latest'}



import logging
import re
import time

from .cromwell_metadata import CromwellMetadata
from .cromwell_rest_api import CromwellRestAPI

logger = logging.getLogger(__name__)


class WorkflowStatusTransition:
    def __init__(self, regex, status_transitions, auto_write_metadata=False):
        """
        Args:
            regex:
                Regular expression to catch workflow's status transition from
                a string (Cromwell's stderr).
                This reg-ex should have only one group which can catch
                workflow's string UUID.
            status_transitions:
                List (or tuple) of possible status transitions.
                Transition is defined by a tuple of previous and next
                statuses where each status is a plain string.
                e.g. [('Submitted', 'Running'),]
                Iterating over this list, only the first valid transition,
                where a previous status is matched, found will be used.
        """
        self._regex = regex
        self._status_transitions = status_transitions
        self._auto_write_metadata = auto_write_metadata

    def parse(self, line, workflow_status_map):
        """
        Args:
            line:
                Line to be parsed to catch status transition.
            workflow_status_map:
                Dict of workflow_id (key) and previus_status (value) pairs.
                This is used to get previous status of a workflow.
                If None then previous status will be ignored.
        Returns:
            workflow_id:
                Workflow's string ID.
            status:
                New status after transition.
            auto_write_metadata:
                For this status transition metadataJSON file should be written
                on workflow's root output directory.
        """
        r = re.findall(self._regex, line)
        if r:
            wf_id = r[0].strip()
            if wf_id in workflow_status_map:
                prev_status = workflow_status_map[wf_id]
            else:
                prev_status = None
            for st1, st2 in self._status_transitions:
                if st1 is None or st1 == prev_status:
                    if st1 != st2:
                        logger.info(
                            'Workflow: id={id}, status={status}'.format(
                                id=wf_id, status=st2
                            )
                        )
                        return wf_id, st2, self._auto_write_metadata
                    break
        return None, None, False


class CromwellWorkflowMonitor:
    """Class constants include several regular expressions to catch
    status changes of workflow/task by Cromwell's STDERR (logging level>=INFO).
    """

    ALL_STATUS_TRANSITIONS = (
        WorkflowStatusTransition(
            regex=r'workflow (\b[0-9a-f]{8}\b-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-\b[0-9a-f]{12}\b) submitted',
            status_transitions=((None, 'Submitted'),),
        ),
        WorkflowStatusTransition(
            regex=r'started WorkflowActor-(\b[0-9a-f]{8}\b-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-\b[0-9a-f]{12}\b)',
            status_transitions=((None, 'Running'),),
        ),
        WorkflowStatusTransition(
            regex=r'Workflow (\b[0-9a-f]{8}\b-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-\b[0-9a-f]{12}\b) failed',
            status_transitions=((None, 'Failed'),),
        ),
        WorkflowStatusTransition(
            regex=r'Abort requested for workflow (\b[0-9a-f]{8}\b-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-\b[0-9a-f]{12}\b)\.',
            status_transitions=((None, 'Aborting'),),
        ),
        WorkflowStatusTransition(
            regex=r'WorkflowActor-(\b[0-9a-f]{8}\b-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-\b[0-9a-f]{12}\b) is in a terminal state',
            status_transitions=(
                ('Failed', 'Failed'),
                ('Aborting', 'Aborted'),
                (None, 'Succeeded'),
            ),
            auto_write_metadata=True,
        ),
        WorkflowStatusTransition(
            regex=r'Workflow actor for (\b[0-9a-f]{8}\b-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-\b[0-9a-f]{12}\b) completed with status',
            status_transitions=(
                ('Failed', 'Failed'),
                ('Aborting', 'Aborted'),
                (None, 'Succeeded'),
            ),
            auto_write_metadata=True,
        ),
    )

    RE_CROMWELL_SERVER_START = r'Cromwell \d+ service started on'
    RE_TASK_START = r'\[UUID\((\b[0-9a-f]{8})\)(.+):(.+):(\d+)\]: job id: (.+)'
    RE_TASK_STATUS_CHANGE = (
        r'\[UUID\((\b[0-9a-f]{8})\)(.+):(.+):(\d+)\]: Status change from (.+) to (.+)'
    )
    RE_TASK_CALL_CACHED = r'\[UUID\((\b[0-9a-f]{8})\)\]: Job results retrieved \(CallCached\): \'(.+)\' \(scatter index: (.+), attempt (\d+)\)'
    RE_SUBWORKFLOW_FOUND = r'(\b[0-9a-f]{8}\b-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-\b[0-9a-f]{12}\b)-SubWorkflowActor-SubWorkflow'

    MAX_RETRY_WRITE_METADATA = 3
    INTERVAL_RETRY_WRITE_METADATA = 10.0
    DEFAULT_SERVER_HOSTNAME = 'localhost'
    DEFAULT_SERVER_PORT = 8000

    def __init__(
        self,
        is_server=False,
        server_hostname=DEFAULT_SERVER_HOSTNAME,
        server_port=DEFAULT_SERVER_PORT,
        embed_subworkflow=False,
        auto_write_metadata=False,
        on_status_change=None,
        on_server_start=None,
    ):
        """Parses STDERR from Cromwell to updates workflow/task information.
        Also, write/update metadata.json on each workflow's root directory.

        Args:
            is_server:
                Cromwell server mode. metadata JSON file update is available
                for server mode only.
                It tries to write/update metadata JSON file on workflow's
                root directory when there is any status change of it.
            metadata_basename:
                Basename for metadata JSON file to be written on each workflow's
                root directory.
            server_hostname:
                Cromwell server hostname for Cromwell REST API.
                This is used to get metadata JSON of a workflow.
            server_hostname:
                Cromwell server port for Cromwell REST API.
                This is used to get metadata JSON of a workflow.
            embed_subworkflow:
                Whenever there is any status change of workflow (or any of its tasks)
                It tries to write/update metadata JSON file on workflow's root.
                For this metadata JSON file, embed subworkflow's metadata JSON in it.
                If this is turned off, then metadata JSON will just have subworkflow's ID.
            auto_write_metadata:
                This is server-only feature. For any change of workflow's status,
                automatically updates metadata JSON file on workflow's root directory.
                metadata JSON is retrieved by communicating with Cromwell server via
                REST API.
            on_status_change:
                Callback function called on any workflow/task status change.
                This should take one parameter (workflow's metadata dict).
                You can parse this dict to get status of workflow and all its task.
                For example,
                    metadata['status']: to get status of a workflow,
                    metadata['id']: to get workflow's ID,
                    metadata['calls']: to get access to list of each task's dict.
                    ...
            on_server_start:
                Callback function called on server start.
                This function should not take parameter.
        """
        self._is_server = is_server

        if self._is_server:
            self._cromwell_rest_api = CromwellRestAPI(
                hostname=server_hostname, port=server_port
            )
        else:
            self._cromwell_rest_api = None

        self._embed_subworkflow = embed_subworkflow
        self._auto_write_metadata = auto_write_metadata
        self._on_status_change = on_status_change
        self._on_server_start = on_server_start

        self._workflow_status_map = dict()
        self._subworkflows = set()
        self._is_server_started = False

    def is_server_started(self):
        return self._is_server_started

    def update(self, stderr):
        """Update workflows by parsing Cromwell's stderr.

        Args:
            stderr:
                stderr from Cromwell.
                Should be a full line (or lines) ending with blackslash n.
        """
        if self._is_server:
            self._update_server_start(stderr)

        updated_workflows, workflows_to_write_metadata = self._update_workflows(stderr)
        self._update_subworkflows(stderr)
        self._update_tasks(stderr)

        for w in workflows_to_write_metadata:
            self._write_metadata(w)

    def _update_server_start(self, stderr):
        if not self._is_server_started:
            for line in stderr.split('\n'):
                r1 = re.findall(CromwellWorkflowMonitor.RE_CROMWELL_SERVER_START, line)
                if r1:
                    self._is_server_started = True
                    if self._on_server_start:
                        self._on_server_start()
                    logger.info('Cromwell server started. Ready to take submissions.')
                    break

    def _update_workflows(self, stderr):
        """Updates workflow status by parsing Cromwell's stderr lines.
        """
        updated_workflows = set()
        workflows_to_write_metadata = set()
        for line in stderr.split('\n'):
            for st_transitions in CromwellWorkflowMonitor.ALL_STATUS_TRANSITIONS:
                workflow_id, status, auto_write_metadata = st_transitions.parse(
                    line, self._workflow_status_map
                )
                if workflow_id:
                    self._workflow_status_map[workflow_id] = status
                    updated_workflows.add(workflow_id)
                    if auto_write_metadata:
                        workflows_to_write_metadata.add(workflow_id)

        return updated_workflows, workflows_to_write_metadata

    def _update_subworkflows(self, stderr):
        for line in stderr.split('\n'):
            r_sub = re.findall(CromwellWorkflowMonitor.RE_SUBWORKFLOW_FOUND, line)
            if r_sub:
                subworkflow_id = r_sub[0]
                if subworkflow_id not in self._subworkflows:
                    logger.info('Subworkflow found: {id}'.format(id=subworkflow_id))
                self._subworkflows.add(subworkflow_id)

    def _update_tasks(self, stderr):
        """Check if workflow's task status changed by parsing Cromwell's stderr lines.
        """
        for line in stderr.split('\n'):
            r_common = None
            r_start = re.findall(CromwellWorkflowMonitor.RE_TASK_START, line)
            if r_start:
                r_common = r_start[0]
                status = 'Started'
                job_id = r_common[4]

            r_callcached = re.findall(CromwellWorkflowMonitor.RE_TASK_CALL_CACHED, line)
            if r_callcached:
                r_common = r_callcached[0]
                status = 'CallCached'
                job_id = None

            r_status_change = re.findall(
                CromwellWorkflowMonitor.RE_TASK_STATUS_CHANGE, line
            )
            if r_status_change:
                r_common = r_status_change[0]
                status = r_common[5]
                job_id = None

            if r_common:
                short_id = r_common[0]
                workflow_id = self._find_workflow_id_from_short_id(short_id)
                task_name = r_common[1]
                shard_idx = r_common[2]
                try:
                    shard_idx = int(shard_idx)
                except ValueError:
                    shard_idx = -1
                retry = int(r_common[3])

                msg = 'Task: id={id}, task={name}:{shard_idx}, retry={retry}, status={status}'.format(
                    id=workflow_id,
                    name=task_name,
                    shard_idx=shard_idx,
                    retry=retry - 1,
                    status=status,
                )
                if job_id:
                    msg += ', job_id={job_id}'.format(job_id=job_id)
                logger.info(msg)

    def _find_workflow_id_from_short_id(self, short_id):
        for w in self._subworkflows.union(set(self._workflow_status_map.keys())):
            if w.startswith(short_id):
                return w

    def _write_metadata(self, workflow_id):
        """Update metadata on Cromwell'e exec root.
        """
        if not self._is_server or not self._auto_write_metadata:
            return
        if workflow_id in self._subworkflows and self._embed_subworkflow:
            logger.debug(
                'Skipped writing metadata JSON file of subworkflow {wf_id}'.format(
                    wf_id=workflow_id
                )
            )
            return
        for trial in range(CromwellWorkflowMonitor.MAX_RETRY_WRITE_METADATA + 1):
            try:
                time.sleep(CromwellWorkflowMonitor.INTERVAL_RETRY_WRITE_METADATA)
                metadata = self._cromwell_rest_api.get_metadata(
                    workflow_ids=[workflow_id],
                    embed_subworkflow=self._embed_subworkflow,
                )[0]
                if self._on_status_change:
                    self._on_status_change(metadata)
                cm = CromwellMetadata(metadata)
                cm.write_on_workflow_root()
            except Exception:
                logger.error(
                    'Failed to retrieve metadata from Cromwell server. '
                    'trial={t}, id={wf_id}'.format(t=trial, wf_id=workflow_id)
                )
                continue
            break



import logging
import os
from copy import deepcopy

from autouri import AutoURI

from .cromwell_backend import (
    BACKEND_AWS,
    BACKEND_GCP,
    BACKEND_SGE,
    CromwellBackendAWS,
    CromwellBackendBase,
    CromwellBackendCommon,
    CromwellBackendDatabase,
    CromwellBackendGCP,
    CromwellBackendLocal,
    CromwellBackendPBS,
    CromwellBackendSGE,
    CromwellBackendSLURM,
)
from .dict_tool import merge_dict
from .hocon_string import HOCONString

logger = logging.getLogger(__name__)


class CaperBackendConf:
    BACKEND_CONF_INCLUDE = 'include required(classpath("application"))'
    BASENAME_BACKEND_CONF = 'backend.conf'

    def __init__(
        self,
        default_backend,
        local_out_dir,
        disable_call_caching=False,
        max_concurrent_workflows=CromwellBackendCommon.DEFAULT_MAX_CONCURRENT_WORKFLOWS,
        memory_retry_error_keys=CromwellBackendCommon.DEFAULT_MEMORY_RETRY_ERROR_KEYS,
        max_concurrent_tasks=CromwellBackendBase.DEFAULT_CONCURRENT_JOB_LIMIT,
        soft_glob_output=False,
        local_hash_strat=CromwellBackendLocal.DEFAULT_LOCAL_HASH_STRAT,
        db=CromwellBackendDatabase.DEFAULT_DB,
        db_timeout=CromwellBackendDatabase.DEFAULT_DB_TIMEOUT_MS,
        mysql_db_ip=CromwellBackendDatabase.DEFAULT_MYSQL_DB_IP,
        mysql_db_port=CromwellBackendDatabase.DEFAULT_MYSQL_DB_PORT,
        mysql_db_user=CromwellBackendDatabase.DEFAULT_MYSQL_DB_USER,
        mysql_db_password=CromwellBackendDatabase.DEFAULT_MYSQL_DB_PASSWORD,
        mysql_db_name=CromwellBackendDatabase.DEFAULT_MYSQL_DB_NAME,
        postgresql_db_ip=CromwellBackendDatabase.DEFAULT_POSTGRESQL_DB_IP,
        postgresql_db_port=CromwellBackendDatabase.DEFAULT_POSTGRESQL_DB_PORT,
        postgresql_db_user=CromwellBackendDatabase.DEFAULT_POSTGRESQL_DB_USER,
        postgresql_db_password=CromwellBackendDatabase.DEFAULT_POSTGRESQL_DB_PASSWORD,
        postgresql_db_name=CromwellBackendDatabase.DEFAULT_POSTGRESQL_DB_NAME,
        file_db=None,
        gcp_prj=None,
        gcp_out_dir=None,
        gcp_call_caching_dup_strat=CromwellBackendGCP.DEFAULT_CALL_CACHING_DUP_STRAT,
        gcp_service_account_key_json=None,
        use_google_cloud_life_sciences=False,
        gcp_region=CromwellBackendGCP.DEFAULT_REGION,
        aws_batch_arn=None,
        aws_region=None,
        aws_out_dir=None,
        aws_call_caching_dup_strat=CromwellBackendAWS.DEFAULT_CALL_CACHING_DUP_STRAT,
        gcp_zones=None,
        slurm_partition=None,
        slurm_account=None,
        slurm_extra_param=None,
        sge_pe=None,
        sge_queue=None,
        sge_extra_param=None,
        pbs_queue=None,
        pbs_extra_param=None,
    ):
        """Initializes the backend conf's stanzas.

        Args:
            default_backend:
                Default backend.
            local_out_dir:
                Output directory for all local backends.
                Define this even if you don't want to run on local backends
                since "Local" is a Cromwell's default backend and it needs this.
            disable_call_caching:
                Disable call-caching (re-using outputs from previous workflows/tasks)
            max_concurrent_workflows:
                Limit for concurrent number of workflows.
            memory_retry_error_keys:
                List of error messages to catch failures due to OOM (out of memory error).
                e.g. ['OutOfMemory', 'Killed']
                If an error occurs caught by these keys, then instance's memory will
                be increased for next retrial by memory_retry_error_multiplier defined
                in workflow options JSON.
            max_concurrent_tasks:
                Limit for concurrent number of tasks for each workflow.
            soft_glob_output:
                Local backends only (Local, sge, pbs, slurm).
                Glob with ln -s instead of hard-linking (ln alone).
                Useful for file-system like beeGFS, which does not allow hard-linking.
            local_hash_strat:
                Local file hashing strategy for call-caching.
            db:
                Metadata DB type. Defauling to use in-memory DB if not defined.
                You may need to define other parameters according to this DB type.
            db_timeout:
                DB connection timeout. Cromwell tries to connect to DB within this timeout.
            mysql_db_ip:
                MySQL DB hostname.
            mysql_db_port:
                MySQL DB port.
            mysql_db_user:
                MySQL DB username.
            mysql_db_password:
                MySQL DB password.
            mysql_db_name:
                MySQL DB name.
            postgresql_db_ip:
                PostgreSQL DB hostname.
            postgresql_db_port:
                PostgreSQL DB port.
            postgresql_db_user:
                PostgreSQL DB user.
            postgresql_db_password:
                PostgreSQL DB password.
            postgresql_db_name:
                PostgreSQL DB name.
            file_db:
                For db == "file". File DB path prefix.
                File DB does not allow multiple connections, which means that
                you cannot run multiple caper run/server with the same file DB.
            gcp_prj:
                Google project name.
            gcp_out_dir:
                Output bucket path for gcp backend. Must start with gs://.
            gcp_call_caching_dup_strat:
                Call-caching duplication strategy for GCP backend.
            gcp_service_account_key_json:
                GCP service account key JSON.
                If defined, service_account scheme will be used instead of application_default
                in Cromwell.
            use_google_cloud_life_sciences:
                Use Google Cloud Life Sciences API.
                This requires only one zone specified in gcp_zones.
                If not specified default zone will be used.
                See Cromwell document:
                    https://cromwell.readthedocs.io/en/stable/backends/Google/.
                Also check supported zones:
                    https://cloud.google.com/life-sciences/docs/concepts/locations
            gcp_region:
                Region for Google Cloud Life Sciences API.
                Ignored if not use_google_cloud_life_sciences.
            aws_batch_arn:
                ARN for AWS Batch.
            aws_region:
                AWS region. Multple regions are not allowed.
            aws_out_dir:
                Output bucket path for aws backend. Must start with s3://.
            aws_call_caching_dup_strat:
                Call-caching duplication strategy for AWS backend.
            gcp_zones:
                List of zones for Google Cloud Genomics API.
                For this and all arguments below this,
                see details in CaperWorkflowOpts.__init__.
                These parameters can be defined either in a backend conf file or
                in a workflow options JSON file.
                One major difference is that the former will also be used as defaults.
            slurm_partition:
            slurm_account:
            slurm_extra_param:
            sge_pe:
            sge_queue:
            sge_extra_param:
            pbs_queue:
            pbs_extra_param:
        """
        self._template = {}

        merge_dict(
            self._template,
            CromwellBackendCommon(
                default_backend=default_backend,
                disable_call_caching=disable_call_caching,
                max_concurrent_workflows=max_concurrent_workflows,
                memory_retry_error_keys=memory_retry_error_keys,
            ),
        )

        merge_dict(
            self._template,
            CromwellBackendDatabase(
                db=db,
                db_timeout=db_timeout,
                file_db=file_db,
                mysql_db_ip=mysql_db_ip,
                mysql_db_port=mysql_db_port,
                mysql_db_user=mysql_db_user,
                mysql_db_password=mysql_db_password,
                mysql_db_name=mysql_db_name,
                postgresql_db_ip=postgresql_db_ip,
                postgresql_db_port=postgresql_db_port,
                postgresql_db_user=postgresql_db_user,
                postgresql_db_password=postgresql_db_password,
                postgresql_db_name=postgresql_db_name,
            ),
        )

        # local backends
        merge_dict(
            self._template,
            CromwellBackendLocal(
                local_out_dir=local_out_dir,
                max_concurrent_tasks=max_concurrent_tasks,
                soft_glob_output=soft_glob_output,
                local_hash_strat=local_hash_strat,
            ),
        )

        merge_dict(
            self._template,
            CromwellBackendSLURM(
                local_out_dir=local_out_dir,
                max_concurrent_tasks=max_concurrent_tasks,
                soft_glob_output=soft_glob_output,
                local_hash_strat=local_hash_strat,
                slurm_partition=slurm_partition,
                slurm_account=slurm_account,
                slurm_extra_param=slurm_extra_param,
            ),
        )

        merge_dict(
            self._template,
            CromwellBackendSGE(
                local_out_dir=local_out_dir,
                max_concurrent_tasks=max_concurrent_tasks,
                soft_glob_output=soft_glob_output,
                local_hash_strat=local_hash_strat,
                sge_pe=sge_pe,
                sge_queue=sge_queue,
                sge_extra_param=sge_extra_param,
            ),
        )

        merge_dict(
            self._template,
            CromwellBackendPBS(
                local_out_dir=local_out_dir,
                max_concurrent_tasks=max_concurrent_tasks,
                soft_glob_output=soft_glob_output,
                local_hash_strat=local_hash_strat,
                pbs_queue=pbs_queue,
                pbs_extra_param=pbs_extra_param,
            ),
        )

        # cloud backends
        if gcp_prj and gcp_out_dir:
            if gcp_service_account_key_json:
                gcp_service_account_key_json = os.path.expanduser(
                    gcp_service_account_key_json
                )
                if not os.path.exists(gcp_service_account_key_json):
                    raise FileNotFoundError(
                        'gcp_service_account_key_json does not exist. f={f}'.format(
                            f=gcp_service_account_key_json
                        )
                    )

            merge_dict(
                self._template,
                CromwellBackendGCP(
                    max_concurrent_tasks=max_concurrent_tasks,
                    gcp_prj=gcp_prj,
                    gcp_out_dir=gcp_out_dir,
                    call_caching_dup_strat=gcp_call_caching_dup_strat,
                    gcp_service_account_key_json=gcp_service_account_key_json,
                    use_google_cloud_life_sciences=use_google_cloud_life_sciences,
                    gcp_region=gcp_region,
                    gcp_zones=gcp_zones,
                ),
            )

        if aws_batch_arn and aws_region and aws_out_dir:
            merge_dict(
                self._template,
                CromwellBackendAWS(
                    max_concurrent_tasks=max_concurrent_tasks,
                    aws_batch_arn=aws_batch_arn,
                    aws_region=aws_region,
                    aws_out_dir=aws_out_dir,
                    call_caching_dup_strat=aws_call_caching_dup_strat,
                ),
            )

        # keep these variables for a backend checking later
        self._sge_pe = sge_pe
        self._gcp_prj = gcp_prj
        self._gcp_out_dir = gcp_out_dir

        self._aws_batch_arn = aws_batch_arn
        self._aws_region = aws_region
        self._aws_out_dir = aws_out_dir

    def create_file(
        self,
        directory,
        backend=None,
        custom_backend_conf=None,
        basename=BASENAME_BACKEND_CONF,
    ):
        """Create a HOCON string and create a backend.conf file.

        Args:
            backend:
                Backend to run a workflow on.
                Default backend will be use if not defined.
            custom_backend_conf:
                User's custom backend conf file to override on
                Caper's auto-generated backend conf.
            basename:
                Basename.
        """
        template = deepcopy(self._template)

        if backend == BACKEND_SGE:
            if self._sge_pe is None:
                raise ValueError(
                    'sge-pe (Sun GridEngine parallel environment) '
                    'is required for backend sge.'
                )
        elif backend == BACKEND_GCP:
            if self._gcp_prj is None:
                raise ValueError(
                    'gcp-prj (Google Cloud Platform project) '
                    'is required for backend gcp.'
                )
            if self._gcp_out_dir is None:
                raise ValueError(
                    'gcp-out-dir (gs:// output bucket path) '
                    'is required for backend gcp.'
                )
        elif backend == BACKEND_AWS:
            if self._aws_batch_arn is None:
                raise ValueError(
                    'aws-batch-arn (ARN for AWS Batch) ' 'is required for backend aws.'
                )
            if self._aws_region is None:
                raise ValueError(
                    'aws-region (AWS region) ' 'is required for backend aws.'
                )
            if self._aws_out_dir is None:
                raise ValueError(
                    'aws-out-dir (s3:// output bucket path) '
                    'is required for backend aws.'
                )

        hocon_s = HOCONString.from_dict(
            template, include=CaperBackendConf.BACKEND_CONF_INCLUDE
        )

        if custom_backend_conf is not None:
            s = AutoURI(custom_backend_conf).read()
            hocon_s.merge(s, update=True)

        final_backend_conf_file = os.path.join(directory, basename)
        AutoURI(final_backend_conf_file).write(str(hocon_s) + '\n')
        return final_backend_conf_file
